unit Unit1;

interface

uses
  SysUtils, Types, Classes, Variants, Graphics, Controls, Forms,
  Dialogs, StdCtrls, Grids, FileCtrl, XPMan, Windows,
  {Engine's units}
  Sys, Objects, Storage, Images, Win32
  ;

type
  TForm1 = class(TForm)
    GroupBox1: TGroupBox;
    List: TListBox;
    GroupBox2: TGroupBox;
    Grid: TStringGrid;
    GroupBox3: TGroupBox;
    Button1: TButton;
    XPMan: TXPManifest;
    Button2: TButton;
    procedure FormCreate(Sender: TObject);
    procedure ListClick(Sender: TObject);
    procedure Button1Click(Sender: TObject);
    procedure Button2Click(Sender: TObject);
  private

    InvFreq: Float;

    procedure ViewInfo(Index: Long);
    procedure UpdateModules;
    procedure TestCMP;
    procedure TestIMG;
    function GetTimeMicroseconds: Float;

  public

  end;


var
  Form1: TForm1;


implementation

{$R *.dfm}


procedure TForm1.FormCreate(Sender: TObject);
var
  Freq: Int64;
begin
  QueryPerformanceFrequency(Freq);
  InvFreq := 1 / Freq;

  Grid.ColWidths[1] := 360;
  Grid.Rows[0].Add('Name');
  Grid.Rows[1].Add('Description');
  Grid.Rows[2].Add('Lib Name');
  Grid.Rows[3].Add('Creator');
  Grid.Rows[4].Add('Language');
  Grid.Rows[5].Add('ID Code');
  Grid.Rows[6].Add('Style');
  Grid.Rows[7].Add('Version');
  Grid.Rows[8].Add('Platform');
  Grid.Rows[9].Add('Requires Lib');
  UpdateModules;
  List.ItemIndex := 0;
  ViewInfo(0);
end;

procedure TForm1.ListClick(Sender: TObject);
begin
  ViewInfo(List.ItemIndex);
end;

procedure TForm1.ViewInfo(Index: Long);
var
  Info: TModuleInfo;
begin
  if (Index > ModuleRegister.ModuleInfos.Count - 1) or
    (List.ItemIndex = -1) then
    Exit;

  Info := ModuleRegister.ModuleInfo[Index].ModuleInfo;

  Grid.Cells[1, 0] := Info.Name;
  Grid.Cells[1, 1] := Info.Description;
  Grid.Cells[1, 2] := Info.LibName;
  Grid.Cells[1, 3] := Info.Creator;
  Grid.Cells[1, 4] := Info.Language;
  Grid.Cells[1, 5] := Info.IDCode;
  Grid.Cells[1, 6] := ModuleStyleToStr(Info.Style);
  Grid.Cells[1, 7] := VersionToStr(Info.Version);
  Grid.Cells[1, 8] := PlatformToStr(Info.Platform);
  Grid.Cells[1, 9] := BoolToStr(Info.RequiresLib, true);
end;

procedure TForm1.Button1Click(Sender: TObject);
var
  Dir: string;
begin
  if not SelectDirectory('Select Directory', '', Dir) then
    Exit;
  ModuleRegister.SearchDirectory(Dir);
  UpdateModules;
end;

procedure TForm1.UpdateModules;
var
  I: Long;
begin
  List.Clear;
  for I := 0 to ModuleRegister.ModuleInfos.Count - 1 do
    List.Items.Add(ModuleRegister.ModuleInfo[I].ModuleInfo.Name);
end;

procedure TForm1.Button2Click(Sender: TObject);
var
  Info: TModuleInfo;
begin
  if (List.ItemIndex > ModuleRegister.ModuleInfos.Count - 1) or
    (List.ItemIndex = -1) then
    Exit;

  Info := ModuleRegister.ModuleInfo[List.ItemIndex].ModuleInfo;

  if Objects.SameIDType('CMP', Info.IDCode) then
    TestCMP;

  if Objects.SameIDType('IMG', Info.IDCode) then
    TestIMG;
end;

function TForm1.GetTimeMicroseconds: Float;
var
  Time: Int64;
begin
  QueryPerformanceCounter(Time);
  Result := 1000000 * InvFreq * Time;
end;

procedure TForm1.TestCMP;
var
  Module: ICompressModule;
  Data, OutData: Pointer;
  InSize, OutSize: Long;
  T1: Float;
begin
  try
    T1 := GetTimeMicroseconds;
    Module :=
      ModuleRegister.CreateModule(ModuleRegister.ModuleInfos[List.ItemIndex]) as
      ICompressModule;
    if Module <> nil then
    begin

      { This is now big problem. Memory managers of supported languages
      behave different so meanwhile
      (until this problem is resolved ? Do you know how ?) we must use something what is the same
      already.}
      Data := Pointer(Windows.LocalAlloc(LMEM_FIXED, 1024 * 1024));
      OutData := nil;
      OutSize := 0;

      Module.CompressBuffer(Data, 1024 * 1024, OutData, OutSize);
      Windows.LocalFree(Long(Data));
      Module.DecompressBuffer(OutData, OutSize, Data, InSize);
      Windows.LocalFree(Long(OutData));
      Windows.LocalFree(Long(Data));
      Module := nil;
    end;
    MsgBox('Module Test Successfull!' + sLineBreak + Format('Time taken: %f us',
      [GetTimeMicroseconds - T1]));
  except
    RaiseError('Module Test Failed! Invalid module: ' + Module.ModuleInfo.Name);
  end;
end;

procedure TForm1.TestIMG;
var
  Module: IImageModule;
  Image: TImage;
  Stream: TMemoryStream;
  T1: Float;
begin
  try
    T1 := GetTimeMicroseconds;
    Module :=
      ModuleRegister.CreateModule(ModuleRegister.ModuleInfos[List.ItemIndex]) as
      IImageModule;
    if Module <> nil then
    begin
      Image := TImage.Create;
      Image.Resize(512, 512);
      Image.ImageFormat := ifA8R8G8B8;
      Stream := TMemoryStream.Create;
      Module.EncodeStream(Image.ImageData, Stream);
      Stream.Position := 0;
      Module.DecodeStream(Stream, Image.ImageData);
      Stream.Free;
      Image.Free;
      Module := nil;
    end;
    MsgBox('Module Test Successfull!' + sLineBreak + Format('Time taken: %f us',
      [GetTimeMicroseconds - T1]));
  except
    RaiseError('Module Test Failed! Invalid module: ' + Module.ModuleInfo.Name);
  end;
end;

end.

