{Misc routines used while debugging on Win32}


unit Win32 platform;

interface

uses
  Sys, Objects, Images,  DIB,WIndows,SysUtils,Graphics,Strings;


type

  TFileListOption  = (flFullNames, flRecursive, flMaskedSubfolders);
  TFileListOptions = set of TFileListOption;

function ImageToBitmap(const Image: TImageData;Level:Long=0): TBitmap;
function BitmapToImage(B: TBitmap): TImageData;
function MsgBox(S: string): Long;overload;
function MsgBox(L: Long): Long;overload;
function AdvBuildFileList(Path: string;  Attr: Integer;
   Files: TStrings; Options: TFileListOptions = [] ;
   SubfoldersMask: string  = '' ): Boolean;
 function StrMatches(Substr, S: string; Index: LongInt = 1 ): Boolean;


implementation

function StrMatches(Substr, S: string; Index: LongInt = 1 ): Boolean;
var
  StringPtr: PChar;
  PatternPtr: PChar;
  StringRes: PChar;
  PatternRes: PChar;

begin
  Result := False;
  StringPtr := PChar(@S[Index]);
  PatternPtr := PChar(SubStr);
  StringRes := nil;
  PatternRes := nil;

  if (S='') or (SubStr='') then
    Exit;

  repeat
    repeat
      case PatternPtr^ of
        #0: begin
              Result:=StringPtr^=#0;
              if Result or (StringRes=nil) or (PatternRes=nil) then
                Exit;

              StringPtr:=StringRes;
              PatternPtr:=PatternRes;
              Break;
            end;
        '*': begin
               inc(PatternPtr);
               PatternRes:=PatternPtr;
               Break;
             end;
        '?': begin
               if StringPtr^=#0 then
                 Exit;
               inc(StringPtr);
               inc(PatternPtr);
             end;
        else
        begin
          if StringPtr^=#0 then
            Exit;
          if StringPtr^<>PatternPtr^ then
          begin
            if (StringRes=nil) or (PatternRes=nil) then
              Exit;
            StringPtr:=StringRes;
            PatternPtr:=PatternRes;
            Break;
          end
          else
          begin
            inc(StringPtr);
            inc(PatternPtr);
          end;
        end;
      end;
    until false;

    repeat
      case PatternPtr^ of
        #0: begin
          Result:=true;
          Exit;
        end;
        '*': begin
          inc(PatternPtr);
          PatternRes:=PatternPtr;
        end;
        '?': begin
          if StringPtr^=#0 then
            Exit;
          inc(StringPtr);
          inc(PatternPtr);
        end;
        else begin
          repeat
            if StringPtr^=#0 then
              Exit;
            if StringPtr^=PatternPtr^ then
              Break;
            inc(StringPtr);
          until false;
          inc(StringPtr);
          StringRes:=StringPtr;
          inc(PatternPtr);
          Break;
        end;
      end;
    until false;
  until false;
end;


// use *.* wildcard for files ( if they have no atribs )
function AdvBuildFileList(Path: string;  Attr: Integer;
   Files: TStrings; Options: TFileListOptions = [] ;
   SubfoldersMask: string  = '' ): Boolean;
  var
  FileMask: string;
  RootDir: string;
  Folders: TStringList;
  CurrentItem: Integer;
  Counter: Integer;
  LocAttr: Integer;

  procedure BuildFolderList;
  var
    FindInfo: TSearchRec;
    Rslt: Integer;
  begin
    Counter := Folders.Count - 1;
    CurrentItem := 0;
    while CurrentItem <= Counter do
    begin
      // searching for subfolders
      Rslt := FindFirst(Folders[CurrentItem] + '*.*', faDirectory, FindInfo);
      try
        while Rslt = 0 do
        begin
          if (FindInfo.Name <> '.') and (FindInfo.Name <> '..') and
            (FindInfo.Attr and faDirectory = faDirectory) then
            Folders.Add(Folders[CurrentItem] + FindInfo.Name + '\');
          Rslt := FindNext(FindInfo);
        end;
      finally
        FindClose(FindInfo);
      end;
      Counter := Folders.Count - 1;
      Inc(CurrentItem);
    end;
  end;

  procedure FillFileList(CurrentCounter: Integer);
  var
    FindInfo: TSearchRec;
    Rslt: Integer;
    CurrentFolder: String;
  begin
    CurrentFolder := Folders[CurrentCounter];
    Rslt := FindFirst(CurrentFolder + FileMask, LocAttr, FindInfo);
    try
      while Rslt = 0 do
      begin
        //if (LocAttr and FindInfo.Attr) =FindInfo.Attr then
          if flFullNames in Options then
            Files.Add(CurrentFolder + FindInfo.Name)
          else
            Files.Add(FindInfo.Name);
        Rslt := FindNext(FindInfo);

      end;
    finally
      FindClose(FindInfo);
    end;
  end;

begin
  FileMask := ExtractFileName(Path);
  RootDir := ExtractFilePath(Path);

  Folders := TStringList.Create;
  Folders.Add(RootDir);

  Files.Clear;

  if Attr = faAnyFile then
    LocAttr := faReadOnly + faHidden + faSysFile + faArchive
  else
    LocAttr := Attr;

  // here's the recursive search for nested folders

  if flRecursive in Options then
    BuildFolderList;

  if Attr<>faDirectory then
  for Counter := 0 to Folders.Count - 1 do
  begin
    if (((flMaskedSubfolders in Options) and (StrMatches(SubfoldersMask, Folders[Counter],1))) or
      (not (flMaskedSubfolders in Options))) then
      FillFileList(Counter);
  end
  else
  Files.AddStrings(Folders);
  Folders.Free;
  Result := True;
end;

function MsgBox(S: string): LongInt;
begin
  Result := MessageBox(GetForegroundWindow, Pchar(S), 'Message', 0);
end;

function MsgBox(L: LongInt): LongInt;
begin
  Result := MessageBox(GetForegroundWindow, Pchar(IntToStr(L)), 'Message', 0);
end;

function ImageToBitmap(const Image: TImageData;Level:Long): TBitmap;
var
  Img: TImageData;
  Data: Pointer;
  Siz: Long;
  D: TDIB;
  B: TBitmap;
begin
  Img := Image;
  with Img,Img.Levels[Level] do
  begin
  FlipYImage(Img,Level);
  AddPadBytes(Img, Data, Siz,Level);
  D := TDIB.Create;
  D.SetSize(Width, Height, BitCount);
//  D.PixelFormat := TDIBPixelFormat(PixelFormat);
  Move(Data^, D.pBits^, Siz);
  FreeMem(Data);

  B := TBitmap.Create;
  B.Width := D.Width;
  B.Height := D.Height;
  B.PixelFormat := pf24bit;

  Move(Palette^,D.ColorTable,PalSize*4) ;
  D.UpdatePalette;
  D.Draw(B.Canvas, B.Canvas.ClipRect);
  D.Free;
  Result := B;
  FlipYImage(Img,Level);
  end;
end;


function BitmapToImage(B: TBitmap): TImageData;
begin

end;


end.

 