{*************************************************************************}
{*   Images                                                              *}
{*   Copyright (C) 2003 Marek Mauder, Vampyre SoftWorks                  *}
{*   Graphic\Images.pas                                                  *}
{*************************************************************************}
{                                                                         }
{ This program is free software; you can redistribute it and/or modify it }
{ under the terms of the GNU General Public License as published by the   }
{ Free Software Foundation; either version 2 of the License, or (at your  }
{ option) any later version. This program is distributed in the hope that }
{ it will be useful, but WITHOUT ANY WARRANTY; without even the implied   }
{ warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See   }
{ the GNU General Public License for more details.                        }
{                                                                         }
{ You should have received a copy of the GNU General Public License       }
{ along with this program; if not, write to the Free Software Foundation, }
{ Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.         }
{                                                                         }
{*************************************************************************}
{  Created On: 21th February 2003  (21.02.2003)                           }
{*************************************************************************}


{This unit contains all that is needed to access/convert/draw/fx/other
image data in various formats as well as color types and functions.}


unit Images;

{DONE -o Galfar: Resolve "flipped" thing in Image Data loading/saving}
{DONE -o Galfar: Add alpha to pixelformat?}
{DONE -o Galfar: Check if setting PixelFormat(0,8,8,8) will work if settings pixels or so}
{DONE -o Galfar: Copy only scanlines in TTGA.LoadRGB !!!}
{DONE -o Galfar: Try if OGl+D3D consume unaligned image and if so, then
rewrite loading to unaligned}
{DONE -o Galfar: Add EncodeFile to ImageModules}
{DONE -o Galfar: Write reading/writing functions for OpenIL to get
support for TStream}
{DONE -o Galfar: Add support for image viewing(debug).}
{DONE -o Galfar: Add support for Dagerfall textures.}
{DONE -o Galfar: Begin high-level images.}
{TODO -o Galfar: Add Vampyre Graphic Imaging format *.vgi.}
{DONE -o Galfar: Add support for SubImages to Image handling routines.}

{DONE -o Galfar: Add format conversions truecolor>8bpp, 8bpp>truecolor, pixel format conversions}
{DONE -o Galfar: Add additional values to TImageFormat}
{TODO -o Galfar: Add support for color keying}
{TODO -o Galfar: Invoke better pixel format routines!! look at glBMP,DevIL,FastDIB,...}
{DONE -o Galfar: Add Clone things to bmp,dagger}
{DONE -o Galfar : Unify Palette functoins to use PPalArray all}

{DONE -o Galfar : Add conversions between intensity modes an other supports}

{TODO -o Galfar : When creating textures try to use reference to
 pointers to data in TImageData rather than copying these data.
 May not work, but can be usefoul.}

{TODO -o Galfar : Add to ImageModule - onelevel formats can save multiplelevel
images as sequence.}
{TODO -o Galfar : ImageEngine and register ImageModules to it}


interface


uses
  Sys, Objects, Storage, SysUtils, Maths;


type

  EImage = class(EStorage);

  {Some color types. Native color format is ARGB (or BGRA somewhere,
  it depends from which side you count bytes)).
  channel:    |A|R|G|B|
  byte:        3 2 1 0  }
  TARGB = ULong;

  TRGB = ULong;

  TBGR = ULong;

  TARGBRec = record
    case Integer of
      0: (Color: TARGB);
      1: (B, G, R, A: Byte);
  end;

  TRGBRec = record
    B, G, R: Byte;
  end;

  TBGRRec = record
    R, G, B: Byte;
  end;

  PBGR = ^TBGR;
  PARGB = ^TARGB;
  PARGBRec = ^TARGBRec;
  PRGBRec = ^TRGBRec;
  PBGRRec = ^TBGRRec;

  TARGBRecArray = array[Word] of TARGBRec;
  PARGBRecArray = ^TARGBRecArray;

  TRGBRecArray = array[Word] of TRGBRec;
  PRGBRecArray = ^TRGBRecArray;


  {Pallette for indexed mode images.
  Alpha will be ignored in the most cases, but ...}
  TPalette = array[0..255] of TARGBRec;
  PPalette = ^TPalette;


  {This custom size palette can be used for some indexed images
  with weird bitcounts or other.}
  TPalArray = array[Word] of TARGBRec;
  PPalArray = ^TPalArray;


  {Pallette for indexed mode images.
  This palette is used for reading/writting from indexed image formats}
  TPalRGB = array[0..255] of TRGBRec;
  PPalRGB = ^TPalRGB;


  {This custom size palette can be used for some indexed images
  with weird bitcounts or other.}
  TPalRGBArray = array[Word] of TRGBRec;
  PPalRGBArray = ^TPalRGBArray;


  {Pixel format is used with bitcounts where channel sizes
  are not whole bytes (like 16bit or 27bit).
  or custom channel sizes are used.
  !!Supports channel sizes max 8bits wide. }
  {TODO -o Galfar : Rewrite}
  TPixelFormat = record
    ABitMask, RBitMask, GBitMask, BBitMask: ULong; // channel mask
    ABitCount, RBitCount, GBitCount, BBitCount: ULong; // bits per channel
    AShift, RShift, GShift, BShift: Long; // channle shift
  end;


  {Defines how many bits are for each channel of pixel data.
  Add new formats to free slots in according group (e.g. new R32G32B32 format
  put after some other RGB formats).
  ifA8R8G8B8 means that A is the most significant byte and B is the least.
  Note: When working with 24,32 bpp images drawing functions assumes
  that pixel format of these images has each channel 8bits wide
  (majority of images will be in these formats -> less checking/determining -> faster). So
  custom channel sizes can be used only for 16/8bit images (meanwhile).}
  TImageFormat = (
    ifUnknown = 0,

    // indexed modes  1..31
    ifIndex8 = 1, // 8bit indexed mode with palette

    // intensity modes (used as single color channels, heightmaps, greyscales, ...)
    //Note: intensity modes except 8bit mode are not supported by any drawing
    // functions which operates with colors in ARGB format. However,
    // SetPixel/GetPixel direct value works
    // When converting 8bit intensity to ARGB format result will be greyscale image.
    // When converting >8bit intensity to ARGB result is uncertain.
    // Howewer, converting functions should provide conversions between
    // various intensity formats.
    // Also resizing and filtering functions should work properly
    // with Intensity formats.
    //  32..63
    ifIntensity8 = 32, // 8bit mode (alpha channel, grey, luminance, heights)
    ifIntensity16 = 33, // 16bit mode (..)
    ifIntensity32 = 34, // 32bit mode (high detail heightmaps )

    // ARGB modes
    // 64..95
    ifR8G8B8 = 64, // std 24bit RGB image

    ifX8R8G8B8 = 65, // 32bit mode, fourth channel is ignored
    ifA8R8G8B8 = 66, // std 32 bit ARGB mode

    ifX1R5G5B5 = 67, // std 15bit RGB mode with 1 ignored bit
    ifR5G6B5 = 68, // std 16bit mode
    ifA1R5G5B5 = 69, // 15bit RGB + 1bit alpha
    ifA4R4G4B4 = 70, // 16bit mode with 4 bits for each channel

    ifR3G3B2 = 71, // 8bit mode
    ifA1R2G3B2 = 72, // 8bit mode with one alpha bit

    //
    ifCustom = 1024
    );


   {Use these intervals in various converting/drawing functions ->
   if you add new ImageFormat which behaves like any other
   (escpecially pixel format using ones like 16bit or 8bit (R3G3B2 ...),
   you don't need to manually add each new format to these functions,
   simply put these intervals instead of exact format.
   Example:  no more this >

   case BitCount of
   8: case ImageFormat of
    ifIndex8: ...
    ifR3G3B2,
    ifA1R2G3B2,
    ifR8:...

    but this >

   case BitCount of
   8: case ImageFormat of
    ifIndex8: ...
    Low(if8bitARGB) .. High(if8bitARGB):...}
  {These need pixel formats}
  if8bitARGBPF = ifR3G3B2..ifA1R2G3B2;
  {These need pixel formats}
  if16bitARGBPF = ifX1R5G5B5..ifA4R4G4B4;
  {These don't need pixel formats}
  if24bitARGB = ifR8G8B8..ifR8G8B8;
  {These don't need pixel formats}
  if32bitARGB = ifX8R8G8B8..ifA8R8G8B8;


  {Structure used by TImageData for holding one level of image.
  Multiple levels can support animation/mipmap/cube-volume textures/ ... }
  TImageLevel = record
    Width: Long; // image width
    Height: Long; // image height
    Size: Long; // actual size of Bits
    Tag: Long; // user data

    Bits: Pointer; // image data
  end;

  PImageLevel = ^TImageLevel;

  TImageLevelArray = array[Word] of TImageLevel;
  PImageLevelArray = ^TImageLevelArray;


  {Structure with only raw image data. Used for image encoders
  and some low-level things.
  These data are:
  ordered:  ARGB ARGB ARGB ...
  Image data in Bits are non-flipped, top scanline is at the beginning of Bits memory
  (image data often stored like this: OGL + D3D textures -> less conversions,
  but in GDI they are stored flipped up).
  No pad bytes or such things are used.
  1 and 4 bpp Images are not supported for saving
  or manipulating. However, they can be loaded and converted to 8 bpp.
  Supports multiple levels which can be used for storing additional frames/mipmaps/other.}
  TImageData = record
    BitCount: Long; // pixel size in bits! this is not sum of channel sizes
    ImageFormat: TImageFormat;
    PalSize: Long; // number of palette entries
    LevelCount: Long; // number of images levels; 1 for standard images
    Tag: Long; // user data

    Palette: PPalArray; // palette
    Levels: PImageLevelArray; // array of images levels(frames, mipmaps);
  end;

  PImageData = ^TImageData;



{*************************************************************************}
{                      Image Encode/Decode Modules                        }
{*************************************************************************}


  {Interface for TImageModule.}
  IImageModule = interface(IStorageModule)
    ['{11479DC7-7508-4637-A94B-4B49101F4761}']

    function GetCanEncode: boolean;
    function GetCanDecode: boolean;
    procedure SetCanEncode(Value: boolean);
    procedure SetCanDecode(Value: boolean);

    function EncodeBuffer(const Image: TImageData; out Data: Pointer; out Size:
      Long): boolean;
    function EncodeStream(const Image: TImageData; Output: TStream): boolean;
    function EncodeFile(const Image: TImageData; const FileName: string):
      boolean;

    function DecodeBuffer(Data: Pointer; Size: Long; var Image: TImageData):
      boolean;
    function DecodeStream(Input: TStream; var Image: TImageData): boolean;
    function DecodeFile(const FileName: string; var Image: TImageData): boolean;

    function TestFormat(Data: Pointer; Size: Long): boolean; overload;
    function TestFormat(Input: TStream): boolean; overload;

    property CanEncode: boolean read GetCanEncode write SetCanEncode;
    property CanDecode: boolean read GetCanDecode write SetCanDecode;

  end;


const

   {ID Code for descendants of TImageModule is 'IMG' so modules can
  have IDcodes like: IMG-BMP, IMG-MNG, ...}
  sImageModuleID = 'IMG';


type

  {Base module for loading/saving various image data formats.
  Important Note!! : When writing Encode/Decode Stream functions
  be sure that your method supports "large streams". This is
  when you call DecodeStream(Mem) and Mem is 20MB stream
  with Position at the  begining of Image data. DecodeStream function
  must be able to read Image from this stream and after reading
  set Position of stream to value exactly after image data.

  If error occurs while encoding/decoding then set stream position
  back to original value so another module for the same image format
  can try luck !!.

  Never trust that Input.Size is everytime exact size of image.  }
  TImageModule = class(TStorageModule, IImageModule)
  protected

    FCanEncode: boolean;
    FCanDecode: boolean;

    function GetCanEncode: boolean;
    function GetCanDecode: boolean;

    procedure SetCanEncode(Value: boolean);
    procedure SetCanDecode(Value: boolean);

    {Call this procedure in EncodeXXX functions. It creates clone of
    Image. If encoder can not save image in Image's original
    format then it is converted (e.g. Converted to 8bpp for Daggerfall modules,
    FlipY for BMP, Swap-Channels for PNG, ... ).
    Clone must be freed after usage!!}
    procedure PrepareImage(const Image: TImageData; out Clone: TImageData);
      virtual;


  public

    constructor Create; override;
    destructor Destroy; override;

    {This func takes valid image data and converts it to specified image
    file format ready to be written into file.}
    function EncodeBuffer(const Image: TImageData; out Data: Pointer; out Size:
      Long): boolean; virtual;

    {This func takes valid image data and converts it to specified image
    file format ready to be written into file.}
    function EncodeStream(const Image: TImageData; Output: TStream):
      boolean; virtual;

    {This function converts image and saves it to external file on disk.
    This default function opens file as TFileStream and calls EncodeStream.}
    function EncodeFile(const Image: TImageData; const FileName: string):
      boolean; virtual;


    {This func converts Data read from image file to valid Image Data
    structure.}
    function DecodeBuffer(Data: Pointer; Size: Long; var Image: TImageData):
      boolean; virtual;

    {This func converts Input stream read from image file to valid Image Data
    structure.}
    function DecodeStream(Input: TStream; var Image: TImageData): boolean;
      virtual;

    {This function loads data from file and converts it to Image Data struct.
    This default function opens file as TFileStream and calls DecodeStream.}
    function DecodeFile(const FileName: string; var Image: TImageData): boolean;
      virtual;


    {This function tests if data are in supported format.}
    function TestFormat(Data: Pointer; Size: Long): boolean; overload; virtual;

    {This function tests if data are in supported format.}
    function TestFormat(Input: TStream): boolean; overload; virtual;


  published

    {Indicates that ImageModule can encode images.}
    property CanEncode: boolean read GetCanEncode write SetCanEncode;

    {Indicates that ImageModule can decode images.}
    property CanDecode: boolean read GetCanDecode write SetCanDecode;

  end;


  {TImageModule class reference.}
  TImageModuleClass = class of TImageModule;


{*************************************************************************}
{                        High level image class                           }
{*************************************************************************}


  {Enum for selecting active resizing filter for TImage.}
  TResizeFilter = (rfPixel, rfBilinear);


  {High level image object which is
  actualy wrapper to TImageData struct. It supports loading/saving various formats,
  drawing (lines, pixels, blits, ....), filtered resizing, pixelformat/bitcount
  changes and other.
  For multiple image formats (mipmaps, anims, ...) you
  must set ActiveLevel property to get access to desired frame.
  Some properties (Width,Height,...) and all drawing functions work with
  ActiveLevel only.}

  {TODO -o Galfar: create TTexture=class(TImage). Call Update to replace device
  texture wih new data. ActiveLevel system can be left as is. Free Image Data
  to get smaller mem usage. Create TBaseTexture, TStaticTexture, TAnimatedTexture.
  Each texture can have multitextures (Levels) created}
  TImage = class(TStorageObject)
  protected

    FActiveLevel: Long;
    FDirect: boolean;
    FResizeFilter: TResizeFilter;

    {Gets various Long properties.}
    function GetPars(Idx: Long): Long; virtual;
    {Sets various Long properties.}
    procedure SetPars(Idx, Value: Long); virtual;

    {Sets new active level.}
    procedure SetActiveLevel(Value: Long); virtual;
    {Converts bitcount to new value. A default ImageFormat for
    each Bitcount is used.}
    procedure ConvertBitCount(Value: Long); virtual;
    {Converts image data to desired format. It also changes bitcount
    if it is required.}
    procedure SetImageFormat(Value: TImageFormat); virtual;

    function GetPixels(X, Y: Long): TARGB;
    function GetPixelsDirect(X, Y: Long): ULong;
    procedure SetPixels(X, Y: Long; Value: TARGB);
    procedure SetPixelsDirect(X, Y: Long; Value: ULong);

    {Initializes underlying TImageData struct.}
    procedure CreateImageData; virtual;
    {Finalizes underlying TImageData struct.}
    procedure FreeImageData; virtual;


  public


    {Raw ImageData structure access.
    Be carefull wih changes !!! They are allowed but no recommended!
    It is not property to get support for functions that operates on
    ImageData e.g. Convert89To658(var Image).}
    ImageData: TImageData;


    {Creates default 16x16x24 one level image.}
    constructor Create; override;

    {Creates TImage from existing TImageData struct.
    It does not copies any data, it uses reference so
    you can not Free source ImageData. It sets Direct property to true
    so destructor will not free ImageData memory.}
    constructor CreateDirect(const Image: TImageData); virtual;

    {Creates TImage from existing TImageData struct.
    It copies all memory data from TImageData so you can
    safelly Free source ImageData.}
    constructor CreateIndirect(const Image: TImageData); virtual;

    destructor Destroy; override;


    {Returns size of memory in bytes occupied by image data stored in
    this object. Returns size of all levels + pal}
    function MemoryUsage: Long; override;

    {Resizes Active Level of image to new dimensions.}
    procedure Resize(AWidth, AHeight: Long); virtual;

    {Resizes All Levels of image to new dimensions.}
    procedure ResizeAll(AWidth, AHeight: Long); virtual;

    {Adds new level of given parameters to image.
    If Activate is true new level will be set as Active Level.
    Returns index of added level.}
    function AddLevel(AWidth, AHeight: Long; Activate: boolean = false): Long;
      overload; virtual;

    {Adds new level with dimension of level 0 to image.
    If Activate is true new level will be set as Active Level.
    Returns index of added level.}
    function AddLevel(Activate: boolean = false): Long;
      overload; virtual;

    {This function removes image level given by ALevel parameter.
    It frees all memory of this level and if ALevel is ActiveLevel
    it sets ActiveLevel to ALevel-1.}
    procedure RemoveLevel(ALevel: Long); virtual;

    {This sets entire alpha channel of image to given value.
    if AllLevels is not set only ActiveLevel is set.
    This function supports 8/16/32bit image formats
    (It sets alpha data even for formats like X1R5G5B5 which
    do not uses this alpha data)
    For ifIndex8 it sets alpha in palette.
    Note: Intensity formats always stores all their data in alpha channel
    so this function would rewrite their data.}
    procedure SetAlphaChannel(Alpha: ULong; AllLevels: boolean = false);


    {Use for direct access to pixels of ActiveLevel. This
    returns raw value stored in image at position of X,Y.}
    property PixelsDirect[X, Y: Long]: ULong read GetPixelsDirect write
    SetPixelsDirect;

    {Use for setting/getting colors of pixels of ActiveLevel. This
    returns color in ARGB format of pixel at X,Y.
    It returns same value as Pixels for 24bit and 32bit images.
    For 16bit images it uses pixelformats and for 8bit it uses palette/pf.}
    property Pixels[X, Y: Long]: TARGB read GetPixels write SetPixels;


  published


    {Active level of image. All drawing/resizing functions operate on
    active level of image.}
    property ActiveLevel: Long index 0 read GetPars write SetPars;

    {Total level count of image.}
    property LevelCount: Long read ImageData.LevelCount;

    {BitCount of image (total size of space for one pixel in bits.
    Not all bits must be used).}
    property BitCount: Long index 1 read GetPars write SetPars;

    {Format of image data.}
    property ImageFormat: TImageFormat read ImageData.ImageFormat write
      SetImageFormat;

    {Image width in pixels (of active level).}
    property Width: Long index 2 read GetPars write SetPars;

    {Image height in pixels (of active level).}
    property Height: Long index 3 read GetPars write SetPars;

    {Indicates that Image was created with CreateDirect construtor.
    It Direct is true and TImage.Destroy is called it will not free
    image data pointers.
    If Direct is false destructor calls FreeImage on ImageData.}
    property Direct: boolean read FDirect write FDirect;

    {Identifies filter which will be used in TImage.Resize functions.}
    property ResizeFilter: TResizeFilter read FResizeFilter write FResizeFilter;

  end;



{*************************************************************************}
{                      ImageData handling functions                           }
{*************************************************************************}


{** Allocation/Freeing **}


{Tests if given Image structure is valid.}
function TestImage(const Image: TImageData): boolean;

{Returns string information about ImageData fields.}
function ImageToStr(const Image: TImageData): string;

{Sets all image pointer to nil and sets all fileds to 0.}
procedure InitImage(var Image: TImageData);

{Creates new image with given attribs.}
procedure NewImage(var Image: TImageData; Width, Height: Long; Format:
  TImageFormat);

{Frees all image pointer and sets all fileds to 0.
It checks if image is valid before freeing pointers.}
procedure FreeImageSafe(var Image: TImageData);

{Frees all image pointer and sets all fileds to 0.
Call only with already initialized ImageData!}
procedure FreeImage(var Image: TImageData);


{** Format Compatibility Converions **}


{Removes padding bytes used by some image formats but not by us.}
procedure RemovePadBytes(var Image: TImageData; Level: Long = 0);

{Adds padding bytes used by some image formats but not by us.
Returns Data&Size of image with added pad bytes.}
procedure AddPadBytes(const Image: TImageData; out Data: Pointer; out Size:
  Long; Level: Long = 0);


{** Misc Color/Data modifiers/getters **}


{Returns pointer to given scanline of image level.}
function GetScanline(const Image: TImageData; Y: Long; Level: Long = 0):
  Pointer;

{Flips scanlines of image data. Default value of Level indicates
that all levels will be flipped.}
procedure FlipYImage(var Image: TImageData; Level: Long = $FFFF);

{Swaps Red and Blue channel of image data.}
procedure SwapChannels(var Image: TImageData; Level: Long = $FFFF);

{This proc creates identical clone of image. All data are copied.}
procedure CloneImage(const Image: TImageData; out Clone: TImageData);


{** Image Format Conversions **}


{Converts 8 bit image to 24 bit RGB image.}
procedure Convert8To24(var Image: TImageData);

{Converts 8 bit to 32 bit ARGB image.}
procedure Convert8To32(var Image: TImageData; DstFormat: TImageFormat =
  ifA8R8G8B8);

{Converts 24bit image to 32 bit ARGB image.}
procedure Convert24To32(var Image: TImageData; DstFormat: TImageFormat =
  ifA8R8G8B8);

{Converts 32bit image to 24 bit image.}
procedure Convert32To24(var Image: TImageData);

{Converts 8bit image to 16 bit image of given format.}
procedure Convert8To16(var Image: TImageData; DstFormat: TImageFormat =
  ifA1R5G5B5);

{Converts 24bit image to 16 bit image of given format.}
procedure Convert24To16(var Image: TImageData; DstFormat: TImageFormat =
  ifA1R5G5B5);

{Converts 24bit image to 16 bit image of given format.}
procedure Convert32To16(var Image: TImageData; DstFormat: TImageFormat =
  ifA1R5G5B5);

{Converts 16bit image to 24bit image.}
procedure Convert16To24(var Image: TImageData);

{Converts 16bit image to 32bit ARGB image.}
procedure Convert16To32(var Image: TImageData; DstFormat: TImageFormat =
  ifA8R8G8B8);

{Converts 16bit image to 16bit image with different format.}
procedure Convert16To16(var Image: TImageData; DstFormat: TImageFormat =
  ifA1R5G5B5);

{Converts 8bit image to 8bit image with different format.}
procedure Convert8To8(var Image: TImageData; DstFormat: TImageFormat =
  ifIndex8);

{Converts 16bit image to 8bit image of given format.}
procedure Convert16To8(var Image: TImageData; DstFormat: TImageFormat =
  ifIndex8);

{Converts 24bit image to 8bit image of given format.}
procedure Convert24To8(var Image: TImageData; DstFormat: TImageFormat =
  ifIndex8);

{Converts 32bit image to 8bit image of given format.}
procedure Convert32To8(var Image: TImageData; DstFormat: TImageFormat =
  ifIndex8);

{Converts image to 8bit ifIntensity8 format using Greayscale conversion.}
procedure ConvertToGreyScale8(var Image: TImageData);

{Converts 8/16/24/32 bit ARGB images to 8bit indexed image.
Uses Halftone palette in R2G4B2 format so result is not very good}
procedure ConvertARGBToIndex(var Image: TImageData);

{Converts 16bit ARGB image to image in custom pixel format.
ImageFormat will be ifCustom. Custom formats are not
supported by any conversion/resize/filter/drawing/... functions.}
procedure Convert16To16Custom(var Image: TImageData; const DstPF: TPixelFormat);

{Converts 16bit custom format image to image with known pixel format.
Custom formats are not supported by any conversion/resize/filter/drawing/... functions.}
procedure Convert16CustomTo16(var Image: TImageData; const SrcPF: TPixelFormat;
  DstFormat: TImageFormat);


{** Image resizing **}


{Set UseScaleRatio to true if you want to
resize additional levels which have different size than Level 0
to scale their dimensions respectivelly to scaling of Level 0.
Example: Image has Lev0: 512x512, Lev1: 256x256. You call
ResizePixels(800,800,Image,2,false) -> All two levels are now
800x800. You call ResizePixels(800,800,Image,2,true) ad
Level 0 is 800x800 and Level 1 is 400x400 (256* 800/512) }

{This functions resize Image to new size. No filtering is used,
only pixels are resized.}
procedure ResizePixels(NewWidth, NewHeight: Long; var Image: TImageData;
  Level: Long = $FFFF; UseScaleRatio: boolean = false);

{Note: Filtered resizing functions are speedest for simple Image Formats
like A8R8G8B8 or IntensityX. 16bit modes which uses Pixel Formats
are very slow and Indexed modes are the slowest of all (uses ClosestColorInPal for
each pixel).}

{This functions resize Image to new size. Uses bilinear
filtering when enlarging image, otherwise calls
ResizePixels and then FilterBilinear.}
procedure ResizeBilinear(NewWidth, NewHeight: Long; var Image: TImageData;
  Level: Long = $FFFF; UseScaleRatio: boolean = false);


{** Image filtering **}


{Note: Filtering functions are speedest for simple Image Formats
like A8R8G8B8 or IntensityX. 16bit modes which uses Pixel Formats
are very slow and Indexed modes are the slowest of all (uses ClosestColorInPal for
each pixel).}

{Applies bilinear (uses 4 nearest pixels) filtering at image.}
procedure FilterBilinear(var Image: TImageData; Level: Long = $FFFF);

{Applies trilinear (uses 9 nearest pixels) filtering at image.}
procedure FilterTrilinear(var Image: TImageData; Level: Long = $FFFF);


{** Image Level functions **}


{Adds new level to image.}
procedure AddLevel(var Image: TImageData);

{Removes level from image.}
procedure RemoveLevel(var Image: TImageData; Level: Long);


{*************************************************************************}
{              Color/PixelFormat/ImageFormat/Palette  Functions           }
{*************************************************************************}


{Color functions
parameters:
ARGB : TARGB - This is color value in fixed ARGB format=>PixelFormat(8,8,8,8).
 One channel has one byte
Color : TARGB - This is color value can be in custom pixel format. TARGB type
 is used only because of 4 byte size
A, R, G, B: Byte - each color channel intensity. Can be 0..255 (0.$FF)
}


{** ARGB Color manipulation **}


{Creates ARGB color value from color channel bytes.}
function ARGB(A, R, G, B: Byte): TARGB;
{Returns color channel bytes from ARGB value.}
procedure GetARGB(ARGB: TARGB; var A, R, G, B: Byte);
{Creates RGB color value from color channel bytes.}
function RGB(R, G, B: Byte): TRGB;
{Creates BGR color value from color channel bytes.}
function BGR(R, G, B: Byte): TBGR;
{Returns alpha channel of ARGB color.}
function GetA(ARGB: TARGB): byte;
{Returns red channel of ARGB color.}
function GetR(ARGB: TARGB): Byte;
{Returns green channel of ARGB color.}
function GetG(ARGB: TARGB): Byte;
{Returns blue channel of ARGB color.}
function GetB(ARGB: TARGB): Byte;

{Creates ARGB record from ARGB color.}
function ARGBRec(ARGB: TARGB): TARGBRec; overload;
{Creates ARGB record from A R G B bytes.}
function ARGBRec(A, R, G, B: Byte): TARGBRec; overload;
{Creates RGB record from ARGB color.}
function RGBRec(ARGB: TARGB): TRGBRec; overload;
{Creates RGB record from ARGB color.}
function RGBRec(R, G, B: Byte): TRGBRec; overload;
{Creates BGR record from ARGB color.}
function BGRRec(ARGB: TARGB): TBGRRec;

{Sets alpha channel of ARGB color.}
function SetA(ARGB: TARGB; A: Byte): TARGB;
{Sets red channel of ARGB color.}
function SetR(ARGB: TARGB; R: Byte): TARGB;
{Sets green channel of ARGB color.}
function SetG(ARGB: TARGB; G: Byte): TARGB;
{Sets blue channel of ARGB color.}
function SetB(ARGB: TARGB; B: Byte): TARGB;


type

  {Types used in color/string conversions.}
  TARGBStrFmt = (afHex, afInt, afBytes, afDescBytes, afFullDesc);

{Returns string representation of Color in hex format.}
function ARGBToStr(ARGB: TARGB): string;
{Returns string representation of Color in selected format.}
function ARGBToStrEx(ARGB: TARGB; Format: TARGBStrFmt): string;


{** Pixel format oprations **}


{Returns pixel format based on given channel sizes.}
function PixelFormat(ABitCount, RBitCount, GBitCount, BBitCount: Long):
  TPixelFormat;
{Returns pixel format based on given channel masks.}
function PixelFormatMask(ABitMask, RBitMask, GBitMask, BBitMask: Long):
  TPixelFormat;

{Returns Color from given bytes encoded by given PF pixel format.
Note: returned color may not be in ARGB format. It depends on pixel format.}
function PFSetARGB(const PF: TPixelFormat; A, R, G, B: Byte): TARGB; overload;
{Returns A,R,G,B color bytes from Color in PF pixel format.}
procedure PFGetARGB(const PF: TPixelFormat; Color: TARGB; var A, R, G, B: Byte);
overload;
{Returns Color from given ARGB color encoded by given PF pixel format.
Note: returned color may not be in ARGB format. It depends on pixel format.}
function PFSetARGB(const PF: TPixelFormat; ARGB: TARGB): TARGB; overload;
{Returns ARGB color value from Color in PF pixel format.}
function PFGetARGB(const PF: TPixelFormat; Color: TARGB): TARGB; overload;

{Converts Pixel format to string info}
function PFToStr(const PF: TPixelFormat): string;


{** Palette operations **}


{This function allocates and sets grey scale palette.
Entries are numer of color entries (256 for 8bit).}
function GreyScalePal(Entries: Long; Alpha: Byte = $FF): PPalArray;

{Creates palette with given bitcount for each channel.}
function HalftonePal(Entries, R, G, B: Long; Alpha: Byte = $FF): PPalArray;

{This function converts RGB pal to ARGB pal.}
function RGBPalToARGBPal(const PalRGB: PPalRGBArray; Entries: Long = 256; Alpha:
  byte = $FF): PPalArray; overload;
function RGBPalToARGBPal(const PalRGB: TPalRGB; Alpha: byte = $FF): TPalette;
overload;

{This function converts ARGB pal to RGB pal.}
function ARGBPalToRGBPal(const PalARGB: PPalArray; Entries: Long = 256):
  PPalRGBArray; overload;
function ARGBPalToRGBPal(const PalARGB: TPalette): TPalRGB; overload;

{This function swaps R and B vaules of palette.}
procedure PalSwapChannels(var PalARGB: PPalArray; Entries: Long = 256);

{Returns color from Pal that is the closest color to Col}
function ClosestColorInPal(Pal: PPalArray; Entries: Long; Col: TARGB): Long;


{** ImageFormat operations **}


{Creates pixel format struct according to given ImageFormat.}
function ImageFormatToPF(Format: TImageFormat): TPixelFormat;

{Returns BitCount of given image format}
function ImageFormatBitCount(Format: TImageFormat): Long;

{Converts ImageFormat enum to its string representation.}
function ImageFormatToStr(Format: TImageFormat): string;



implementation


uses
  {Internal image modules that are automatically registered}
  BmpModule, TgaModule, PngModule, JpgModule, Dagger;


{*************************************************************************}
{                         TImageModule Methods                           }
{*************************************************************************}


constructor TImageModule.Create;
begin
  inherited Create;
  FCanEncode := true;
  FCanDecode := true;
end;

destructor TImageModule.Destroy;
begin

  inherited Destroy;
end;

function TImageModule.GetCanEncode: boolean;
begin
  Result := FCanEncode;
end;

function TImageModule.GetCanDecode: boolean;
begin
  Result := FCanDecode;
end;

procedure TImageModule.SetCanEncode(Value: boolean);
begin
  FCanEncode := Value;
end;

procedure TImageModule.SetCanDecode(Value: boolean);
begin
  FCanDecode := Value;
end;

procedure TImageModule.PrepareImage(const Image: TImageData; out Clone:
  TImageData);
begin
  CloneImage(Image, Clone);
end;

function TImageModule.EncodeBuffer(const Image: TImageData;
  out Data: Pointer; out Size: Long): boolean;
begin
  Result := false;
end;

function TImageModule.EncodeStream(const Image: TImageData;
  Output: TStream): boolean;
begin
  Result := false;
end;

function TImageModule.EncodeFile(const Image: TImageData;
  const FileName: string): boolean;
var
  F: TFileStream;

begin
  Result := false;
  F := nil;
  try
    F := TFileStream.Create(FileName, fmCreate);
  except
    on E1: EFCreateError do
      RaiseError(ErrString(Self, 'EncodeFile', E1.Message), EFCreateError);
  end;
  try
    try
    // calls EncodeStream function rather
      Result := EncodeStream(Image, F);
    except
      RaiseError(ErrString(Self, 'EncodeFile', 'Error when encoding file "' +
        FileName + '" !'), EImage);
    end;
  finally
    F.Free;
  end;
end;

function TImageModule.DecodeBuffer(Data: Pointer; Size: Long;
  var Image: TImageData): boolean;
begin
  Result := false;
end;

function TImageModule.DecodeStream(Input: TStream;
  var Image: TImageData): boolean;
begin
  Result := false;
end;

function TImageModule.DecodeFile(const FileName: string;
  var Image: TImageData): boolean;
var
  F: TFileStream;

begin
  Result := false;
  F := nil;
  try
    F := TFileStream.Create(FileName, fmOpenRead);
  except
    on E1: EFOpenError do
      RaiseError(ErrString(Self, 'DecodeFile', E1.Message), EFOpenError);
  end;
  try
    try
    // calls DecodeStream function rather
      Result := DecodeStream(F, Image);
    except
      RaiseError(ErrString(Self, 'DecodeFile', 'Error when decoding file "' +
        FileName + '" !'), EImage);
    end;
  finally
    F.Free;
  end;
end;

function TImageModule.TestFormat(Data: Pointer; Size: Long): boolean;
begin
  Result := false;
end;

function TImageModule.TestFormat(Input: TStream): boolean;
begin
  Result := false;
end;


{*************************************************************************}
{                         High level image class                          }
{*************************************************************************}


constructor TImage.Create;
begin
  inherited Create;
  FDirect := false;
  FActiveLevel := 0;
  FResizeFilter := rfPixel;
  if not TestImage(ImageData) then
    CreateImageData;
end;

constructor TImage.CreateDirect(const Image: TImageData);
begin
  ImageData := Image;
  Create;
  FDirect := true;
end;

constructor TImage.CreateIndirect(const Image: TImageData);
begin
  CloneImage(Image, ImageData);
  Create;
  FDirect := false;
end;

destructor TImage.Destroy;
begin
  if not FDirect then
    FreeImageData;
  inherited Destroy;
end;

function TImage.GetPars(Idx: Long): Long;
begin
  Result := 0;
  case Idx of
    0: Result := FActiveLevel;
    1: Result := ImageData.BitCount;
    2: Result := ImageData.Levels[FActiveLevel].Width;
    3: Result := ImageData.Levels[FActiveLevel].Height;
  end;
end;

procedure TImage.SetPars(Idx, Value: Long);
begin
  case Idx of
    0: SetActiveLevel(Value);
    1: ConvertBitCount(Value);
    2: Resize(Value, Height);
    3: Resize(Width, Value);
  end;
end;

procedure TImage.ConvertBitCount(Value: Long);
begin
  if ImageData.BitCount <> Value then
    case Value of
      8: SetImageFormat(ifIndex8);
      16: SetImageFormat(ifA1R5G5B5);
      24: SetImageFormat(ifR8G8B8);
      32: SetImageFormat(ifA8R8G8B8);
    end;
end;

procedure TImage.SetActiveLevel(Value: Long);
begin
  FActiveLevel := Value;
end;

procedure TImage.SetImageFormat(Value: TImageFormat);
var
  SrcBpp, DstBpp: Long;
  OldFmt: TImageFormat;
begin
  if ImageData.ImageFormat = Value then
    Exit;

  SrcBpp := ImageFormatBitCount(ImageData.ImageFormat);
  DstBpp := ImageFormatBitCount(Value);
  OldFmt := ImageData.ImageFormat;

  try
    case SrcBpp of
      8: case DstBpp of
          8: Convert8To8(ImageData, Value);
          16: Convert8To16(ImageData, Value);
          24: Convert8To24(ImageData);
          32: Convert8To32(ImageData);
        end;
      16: case DstBpp of
          8: Convert16To8(ImageData, Value);
          16: Convert16To16(ImageData, Value);
          24: Convert16To24(ImageData);
          32: Convert16To32(ImageData);
        end;
      24: case DstBpp of
          8: Convert24To8(ImageData, Value);
          16: Convert24To16(ImageData, Value);
          32: Convert24To32(ImageData);
        end;
      32: case DstBpp of
          8: Convert32To8(ImageData, Value);
          16: Convert32To16(ImageData, Value);
          24: Convert32To24(ImageData);
          32: // it is done by SetAlphaChannel(0) below
        end;
    end;

    // check alpha channels
    if (ImageFormatToPF(OldFmt).ABitCount > 0) and
      (ImageFormatToPF(Value).ABitCount = 0) then
      SetAlphaChannel(0);
    if (ImageFormatToPF(OldFmt).ABitCount = 0) and
      (ImageFormatToPF(Value).ABitCount > 0)
      and not (OldFmt in [ifIndex8..TImageFormat(31)]) then
      SetAlphaChannel($FF);

  except
    RaiseError(ErrString(Self, 'SetImageFormat',
      'Error when converting image to another data format!' + sLineBreak +
      'Converting from "' + ImageFormatToStr(OldFmt) + '" to "' +
      ImageFormatToStr(Value) + '".'), EImage);
  end;
end;

function TImage.GetPixelsDirect(X, Y: Long): ULong;
begin
  Result := 0;
  with ImageData, ImageData.Levels[FActiveLevel] do
    case BitCount of
      8: Result := PByteArray(Long(Bits) + Y * Width)[X];
     // or use PWord(Long(Bits)+Y*Width*2+X*2); , but there is additional mult
      16: Result := PWordArray(Long(Bits) + Y * (Width shl 1))[X];
      24: with PRGBRecArray(Long(Bits) + Y * Width * 3)[X] do
          Result := (R shl 16) or (G shl 8) or B;
      32: Result := PULongArray(Long(Bits) + Y * (Width shl 2))[X];
    end;
end;

function TImage.GetPixels(X, Y: Long): TARGB;
var
  B: Byte;
  PF: TPixelFormat;
begin
  Result := 0;
  with ImageData, ImageData.Levels[FActiveLevel] do
    case BitCount of
      8: case ImageFormat of
          ifIndex8: Result := Palette[PByteArray(Long(Bits) + Y *
              Width)[X]].Color;
          ifIntensity8:
            begin
              B := PByteArray(Long(Bits) + Y * Width)[X];
              Result := (B shl 24);
            end;
          Low(if8bitARGBPF)..High(if8bitARGBPF):
            begin
              PF := ImageFormatToPF(ImageFormat);
              Result := PFGetARGB(PF, PByteArray(Long(Bits) + Y * Width)[X]);
            end;
        end;
      16: case ImageFormat of
          Low(if16bitARGBPF)..High(if16bitARGBPF):
            begin
              PF := ImageFormatToPF(ImageFormat);
              Result := PFGetARGB(PF, PWordArray(Long(Bits) + Y * (Width shl
                1))[X]);
            end;
          ifIntensity16:
            Result := PWordArray(Long(Bits) + Y * (Width shl 1))[X];
        end;
      24: with PRGBRecArray(Long(Bits) + Y * Width * 3)[X] do
          Result := (R shl 16) or (G shl 8) or B;
      32: Result := PULongArray(Long(Bits) + Y * (Width shl 2))[X];
    end;
end;

procedure TImage.SetPixelsDirect(X, Y: Long; Value: ULong);
begin
  with ImageData, ImageData.Levels[FActiveLevel] do
    case BitCount of
      8: PByteArray(Long(Bits) + Y * Width)[X] := Value;
      16: PWordArray(Long(Bits) + Y * (Width shl 1))[X] := Value;
      24: with PRGBRecArray(Long(Bits) + Y * Width * 3)[X] do
        begin
          R := Byte(Value shr 16);
          G := Byte(Value shr 8);
          B := Byte(Value);
        end;
      32: PULongArray(Long(Bits) + Y * (Width shl 2))[X] := Value;
    end;
end;

procedure TImage.SetPixels(X, Y: Long; Value: TARGB);
var
  PF: TPixelFormat;
begin
  with ImageData, ImageData.Levels[FActiveLevel] do
    case BitCount of
      8: case ImageFormat of
          ifIndex8: PByteArray(Long(Bits) + Y * Width)[X] :=
            ClosestColorInPal(Palette, PalSize, Value);
          // we assume that ARGB color for intensity has same value of each channel
          ifIntensity8: PByteArray(Long(Bits) + Y * Width)[X] := Value;
          Low(if8bitARGBPF)..High(if8bitARGBPF):
            begin
              PF := ImageFormatToPF(ImageFormat);
              PByteArray(Long(Bits) + Y * Width)[X] := PFSetARGB(PF, Value);
            end;
        end;
      16: case ImageFormat of
          Low(if16bitARGBPF)..High(if16bitARGBPF):
            begin
              PF := ImageFormatToPF(ImageFormat);
              PWordArray(Long(Bits) + Y * (Width shl 1))[X] := PFSetARGB(PF,
                Value);
            end;
          ifIntensity16:
            PWordArray(Long(Bits) + Y * (Width shl 1))[X] := Value;
        end;
      24: with PRGBRecArray(Long(Bits) + Y * Width * 3)[X] do
        begin
          R := Byte(Value shr 16);
          G := Byte(Value shr 8);
          B := Byte(Value);
        end;
      32: PULongArray(Long(Bits) + Y * (Width shl 2))[X] := Value;
    end;
end;

procedure TImage.CreateImageData;
begin
  FreeImageData;
  // create default starting image
  InitImage(ImageData);
  with ImageData do
  begin
    BitCount := 24;
    ImageFormat := ifR8G8B8;
    Images.AddLevel(ImageData);
    Levels[0].Width := 16;
    Levels[0].Height := 16;
    Levels[0].Size := 16 * 16 * 3;
    GetMem(Levels[0].Bits, 16 * 16 * 3);
  end;
  Assert(TestImage(ImageData), ErrString(Self, 'CreateImageData',
    'Unknown error when creating' +
    ' default image! This should never happen!'));
end;

procedure TImage.FreeImageData;
begin
  FreeImageSafe(ImageData);
end;

function TImage.MemoryUsage: Long;
var
  I: Long;
begin
  Result := ImageData.PalSize * SizeOf(TARGBRec);
  try
    for I := 0 to ImageData.LevelCount - 1 do
      Inc(Result, ImageData.Levels[I].Size);
  except
    Result := 0;
  end;
end;

procedure TImage.Resize(AWidth, AHeight: Long);
begin
 {TODO -o Galfar: Add code here resize}
  case FResizeFilter of
    rfPixel: ResizePixels(AWidth, AHeight, ImageData, ActiveLevel, false);
    rfBilinear: ResizeBilinear(AWidth, AHeight, ImageData, ActiveLevel, false);
  end;
end;

procedure TImage.ResizeAll(AWidth, AHeight: Long);
var
  I: Long;
  OldLevel: Long;
begin
  OldLevel := ActiveLevel;
  for I := 0 to LevelCount - 1 do
  begin
    ActiveLevel := I;
    Resize(AWidth, AHeight);
  end;
  ActiveLevel := OldLevel;
end;

function TImage.AddLevel(AWidth, AHeight: Long; Activate: boolean): Long;
begin
  Result := ImageData.LevelCount;
  with ImageData do
  try
    Images.AddLevel(ImageData);
    with ImageData.Levels[Result] do
    begin
      Width := AWidth;
      Height := AHeight;
      Size := Width * Height * (BitCount shr 3);
      GetMemFill(Bits, Size);
    end;
  except
    RaiseError(ErrString(Self, 'AddLevel',
      'Error when adding new level to image'), EImage);
  end;
  if Activate then
    ActiveLevel := Result;
end;

function TImage.AddLevel(Activate: boolean): Long;
begin
  Assert(ImageData.LevelCount > 0, ErrString(Self, 'AddLevel', 'Can not create'
    + ' new level with dimensions of level 0! Level 0 does not exist!'));
  Result := AddLevel(ImageData.Levels[0].Width, ImageData.Levels[0].Height,
    Activate);
end;

procedure TImage.RemoveLevel(ALevel: Long);
begin
  Images.RemoveLevel(ImageData, ALevel);
  if ActiveLevel = ALevel then
    ActiveLevel := Max(0, ALevel - 1);
end;

procedure TImage.SetAlphaChannel(Alpha: ULong; AllLevels: boolean);
var
  RGBSize, I, Min, Max, X, Y: Long;
  CA: PARGBRec;
  U: PULong;
  PF, NewPF: TPixelFormat;
  W: PWord;
  C: TARGB;
  B: PByte;

begin
  Max := Iff(AllLevels, ImageData.LevelCount - 1, ActiveLevel);
  Min := Iff(AllLevels, 0, ActiveLevel);
  case ImageFormatBitCount(ImageData.ImageFormat) of
    8: case ImageData.ImageFormat of
        ifIndex8:
          for I := 0 to ImageData.PalSize - 1 do
            ImageData.Palette[I].A := Alpha;
        ifIntensity8:
          for I := Min to Max do
          begin
            B := ImageData.Levels[I].Bits;
            for X := 0 to ImageData.Levels[I].Width - 1 do
              for Y := 0 to ImageData.Levels[I].Height - 1 do
              begin
                B^ := Alpha;
                Inc(B);
              end;
          end;
        Low(if8bitARGBPF)..High(if8bitARGBPF):
          begin
            PF := ImageFormatToPF(ImageData.ImageFormat);
            RGBSize := PF.RBitCount + PF.GBitCount + PF.BBitCount;
            if RGBSize = 8 then
              Exit; // there is no space for alpha channel data

            if PF.ABitCount = 0 then
              NewPF := PixelFormat(8 - RGBSize, PF.RBitCount, PF.GBitCount,
                PF.BBitCount)
            else
              NewPF := PF;

            for I := Min to Max do
            begin
              B := ImageData.Levels[I].Bits;
              for X := 0 to ImageData.Levels[I].Width - 1 do
                for Y := 0 to ImageData.Levels[I].Height - 1 do
                begin
                  C := PFGetARGB(NewPF, B^);
                  B^ := PFSetARGB(NewPF, SetA(C, Alpha));
                  Inc(B);
                end;
            end;
          end;
      end;
    16: case ImageData.ImageFormat of
        ifIntensity16:
          for I := Min to Max do
          begin
            W := ImageData.Levels[I].Bits;
            for X := 0 to ImageData.Levels[I].Width - 1 do
              for Y := 0 to ImageData.Levels[I].Height - 1 do
              begin
                W^ := Alpha;
                Inc(W);
              end;
          end;
        Low(if16bitARGBPF)..High(if16bitARGBPF):
          begin
            PF := ImageFormatToPF(ImageData.ImageFormat);
            RGBSize := PF.RBitCount + PF.GBitCount + PF.BBitCount;
            if RGBSize = 16 then
              Exit; // there is no space for alpha channel data

            if PF.ABitCount = 0 then
              NewPF := PixelFormat(16 - RGBSize, PF.RBitCount, PF.GBitCount,
                PF.BBitCount)
            else
              NewPF := PF;

            for I := Min to Max do
            begin
              W := ImageData.Levels[I].Bits;
              for X := 0 to ImageData.Levels[I].Width - 1 do
                for Y := 0 to ImageData.Levels[I].Height - 1 do
                begin
                  C := PFGetARGB(NewPF, W^);
                  W^ := PFSetARGB(NewPF, SetA(C, Alpha));
                  Inc(W);
                end;
            end;
          end;
      end;
    32: case ImageData.ImageFormat of
        Low(if32bitARGB)..High(if32bitARGB):
          for I := Min to Max do
          begin
            CA := ImageData.Levels[I].Bits;
            for X := 0 to ImageData.Levels[I].Width - 1 do
              for Y := 0 to ImageData.Levels[I].Height - 1 do
              begin
                CA.A := Alpha;
                Inc(CA);
              end;
          end;
        ifIntensity32:
          for I := Min to Max do
          begin
            U := ImageData.Levels[I].Bits;
            for X := 0 to ImageData.Levels[I].Width - 1 do
              for Y := 0 to ImageData.Levels[I].Height - 1 do
              begin
                U^ := Alpha;
                Inc(U);
              end;
          end;
      end;
  end;
end;


{*************************************************************************}
{                      Image handling functions                           }
{*************************************************************************}


function TestImage(const Image: TImageData): boolean;
begin
  Result := false;
  if Image.LevelCount > 0 then
    with Image, Image.Levels[0] do
    begin
      if ImageFormat = ifUnknown then
        Exit;

      if (Width < 1) or (Height < 1) or (Size = 0) or (BitCount < 1) then
        Exit;

      if (PalSize > 0) and (Palette = nil) then
        Exit;

      if (Bits = nil) then
        Exit;

      if (Width * Height * (BitCount shr 3) <> Size) and (BitCount >= 8) then
        Exit;

      Result := true;
    end;
end;

function ImageToStr(const Image: TImageData): string;
var
  S: string;
begin
  with Image do
  begin
    S := Format('Bpp:  %d, Image Fomat: (' + ImageFormatToStr(Image.ImageFormat)
      + '), Levels: %d, PalSize: %d, Palette: ',
      [BitCount, LevelCount, PalSize]);
    if Palette = nil then
      S := S + 'nil,  '
    else
      S := S + Format('$%.8x,  ', [ULong(Palette)]);
    if LevelCount > 0 then
    begin
      S := S + Format('Level 0:(W: %d  H: %d  Size: %d  Bits: ',
        [Levels[0].Width, Levels[0].Height, Levels[0].Size]);
      if Levels[0].Bits = nil then
        S := S + 'nil)  '
      else
        S := S + Format('$%.8x)', [ULong(Levels[0].Bits)]);
    end
    else
      S := 'Level 0: Not present'
  end;
  Result := S;
end;

procedure InitImage(var Image: TImageData);
begin
  FillChar(Image, SizeOf(Image), 0);
end;

procedure NewImage(var Image: TImageData; Width, Height: Long; Format:
  TImageFormat);
begin
  FreeImageSafe(Image);
  AddLevel(Image);
  Image.ImageFormat := Format;
  Image.BitCount := ImageFormatBitCount(Format);
  if Format = ifIndex8 then
  begin
    Image.PalSize := 256;
    Image.Palette := HalftonePal(256, 2, 4, 2, $FF);
  end;
  Image.Levels[0].Width := Width;
  Image.Levels[0].Height := Height;
  Image.Levels[0].Size := Width * Height * (Image.BitCount shr 3);
  GetMem(Image.Levels[0].Bits, Image.Levels[0].Size);
end;

procedure FreeImageSafe(var Image: TImageData);
var
  I: Long;
begin
  // checks image validity
  with Image do
    if (LevelCount > 0) and (BitCount in [8, 16, 24, 32]) then
    try
      if PalSize > 0 then
        FreeMemNil(Palette);
      for I := 0 to LevelCount - 1 do
        FreeMemNil(Levels[I].Bits);
      if LevelCount > 0 then
        FreeMemNil(Levels);
    except
    // pointers are not nil, but they are not valid
    end;
  InitImage(Image);
end;

procedure FreeImage(var Image: TImageData);
var
  I: Long;
begin
  with Image do
  try
    if PalSize > 0 then
      FreeMemNil(Palette);
    for I := 0 to LevelCount - 1 do
      FreeMemNil(Levels[I].Bits);
    if LevelCount > 0 then
      FreeMemNil(Levels);
  except
  // pointers are not nil, but they are not valid
  end;
  InitImage(Image);
end;

procedure RemovePadBytes(var Image: TImageData; Level: Long);
var
  I: Long;
  Data: Pointer;
  W, WB: Long;

begin
  with Image, Image.Levels[Level] do
  begin
    W := Width * (BitCount shr 3);
    I := Height * W;
    // already removed or invalid
    if (Size = I) or (Bits = nil) then
      Exit;
    Size := I;
    WB := (((Width * BitCount) + 31) shr 5) * 4;
    GetMem(Data, Size);
    try
      for I := 0 to Height - 1 do
        Move(Pointer(Long(Bits) + I * WB)^, Pointer(Long(Data) + I * W)^, W);
    finally
      FreeMemNil(Bits);
      Bits := Data;
    end;
  end;
end;

procedure AddPadBytes(const Image: TImageData; out Data: Pointer; out Size:
  Long; Level: Long = 0);
var
  I, W, WB: Long;

begin
  WB := (((Image.Levels[Level].Width * Image.BitCount) + 31) shr 5) * 4;
  W := Image.Levels[Level].Width * (Image.BitCount shr 3);

  // this is one of the most weird thing I've ever seen!!
  // if you set Size it also changes Image.Levels[Level].Size !!!!
  I := Image.Levels[Level].Size;
  Size := Image.Levels[Level].Height * WB;
  Image.Levels[Level].Size := I;
  if (Image.Levels[Level].Bits = nil) then
    Exit;
  GetMem(Data, Size);
  for I := 0 to Image.Levels[Level].Height - 1 do
    Move(Pointer(Long(Image.Levels[Level].Bits) + I * W)^, Pointer(Long(Data) + I
      * WB)^, W);
end;

procedure FlipYImage(var Image: TImageData; Level: Long = $FFFF);
var
  P1, P2, Buff: Pointer;
  Mi, Ma, W, I, J: Long;

begin
  Mi := Iff(Level <> $FFFF, Level, 0);
  Ma := Iff(Level <> $FFFF, Level, Image.LevelCount - 1);
  for J := Mi to Ma do
    with Image, Image.Levels[J] do
    begin
      W := Width * (BitCount shr 3);
      GetMem(Buff, W);
      try
        for I := 0 to Height shr 1 - 1 do
        begin
          P1 := Pointer(Long(Bits) + I * W);
          P2 := Pointer(Long(Bits) + (Height - I - 1) * W);
          Move(P1^, Buff^, W);
          Move(P2^, P1^, W);
          Move(Buff^, P2^, W);
        end;
      finally
        FreeMemNil(Buff);
      end;
    end;
end;

procedure SwapChannels(var Image: TImageData; Level: Long = $FFFF);
var
  ColA: PARGBRec;
  Col: PRGBRec;
  C: TARGB;
  PF: TPixelFormat;
  B: Byte;
  PW: PWord;
  PB: PByte;
  Mi, Ma, X, Y, I: Long;
begin
  Mi := Iff(Level <> $FFFF, Level, 0);
  Ma := Iff(Level <> $FFFF, Level, Image.LevelCount - 1);
  for I := Mi to Ma do
    with Image, Image.Levels[I] do
      case BitCount of
        8: case ImageFormat of
            ifIndex8:
              for X := 0 to PalSize - 1 do
              begin
                B := Palette[X].B;
                Palette[X].B := Palette[X].R;
                Palette[X].R := B;
              end;
            ifIntensity8: ; //nor R nor B here
            Low(if8bitARGBPF)..High(if8bitARGBPF):
              begin
                PF := ImageFormatToPF(ImageFormat);
                PB := Bits;
                for X := 0 to Width - 1 do
                  for Y := 0 to Height - 1 do
                  begin
                    C := PFGetARGB(PF, PB^);
                    PB^ := PFSetARGB(PF, $FF, C, C shr 8, C shr 16);
                    Inc(PB);
                  end;
              end;
          end;
        16: case ImageFormat of
            ifIntensity16: ;
            Low(if8bitARGBPF)..High(if8bitARGBPF):
              begin
                PF := ImageFormatToPF(ImageFormat);
                PW := Bits;
                for X := 0 to Width - 1 do
                  for Y := 0 to Height - 1 do
                  begin
                    C := PFGetARGB(PF, PW^);
                    PW^ := PFSetARGB(PF, C shr 24, C, C shr 8, C shr 16);
                    Inc(PW);
                  end;
              end;
          end;
        24:
          begin
            Col := Bits;
            for X := 0 to Width - 1 do
              for Y := 0 to Height - 1 do
              begin
                B := Col.R;
                Col.R := Col.B;
                Col.B := b;
                Inc(Col);
              end;
          end;
        32: case ImageFormat of
            ifIntensity32: ;
            Low(if32bitARGB)..High(if32bitARGB):
              begin
                ColA := Bits;
                for X := 0 to Width - 1 do
                  for Y := 0 to Height - 1 do
                  begin
                    B := ColA.R;
                    ColA.R := ColA.B;
                    ColA.B := B;
                    Inc(ColA);
                  end;
              end;
          end;
      end;
end;

procedure CloneImage(const Image: TImageData; out Clone: TImageData);
var
  I: Long;

begin
  InitImage(Clone);
  if TestImage(Image) then
    with Image do
    begin
      Clone.BitCount := BitCount;
      Clone.ImageFormat := ImageFormat;
      Clone.PalSize := PalSize;
      Clone.LevelCount := LevelCount;
      Clone.Tag := Tag;

      if PalSize > 0 then
      begin
        GetMemFill(Clone.Palette, PalSize * SizeOf(TARGBRec));
        Move(Palette^, Clone.Palette^, PalSize * SizeOf(TARGBRec));
      end;

      GetMem(Clone.Levels, LevelCount * SizeOf(TImageLevel));
      for I := 0 to LevelCount - 1 do
        with Image.Levels[I], Clone do
        begin
          Levels[I].Width := Width;
          Levels[I].Height := Height;
          Levels[I].Size := Size;
          Levels[I].Tag := Tag;
          GetMem(Levels[I].Bits, Size);
          Move(Bits^, Levels[I].Bits^, Size);
        end;
    end;
end;


procedure Convert8To24(var Image: TImageData);
var
  I, X, Y: Long;
  Data: Pointer;
  C: PRGBRec;
  PB: PByte;
  A, R, G, B: Byte;
  Pal: PPalArray;
  PF: TPixelFormat;
begin
  if not TestImage(Image) then
    Exit;

  Pal := nil;
  case Image.ImageFormat of
    ifIndex8, ifIntensity8:
      begin
        if Image.Palette = nil then
          Pal := GreyScalePal(256)
        else
          Pal := Image.Palette;
        for I := 0 to Image.LevelCount - 1 do
          with Image, Image.Levels[I] do
          begin
            GetMem(Data, Width * Height * 3);
            C := Data;
            Size := Width * Height * 3;
            PB := Bits;
            for X := 0 to Width - 1 do
              for Y := 0 to Height - 1 do
              begin
                C.R := Pal[PB^].R;
                C.G := Pal[PB^].G;
                C.B := Pal[PB^].B;
                Inc(C);
                Inc(PB);
              end;
            FreeMemNil(Bits);
            Bits := Data;
          end;
      end;
    Low(if8bitARGBPF)..High(if8bitARGBPF):
      begin
        PF := ImageFormatToPF(Image.ImageFormat);
        for I := 0 to Image.LevelCount - 1 do
          with Image, Image.Levels[I] do
          begin
            PB := Bits;
            GetMem(Data, Width * Height * 3);
            C := Data;
            Size := Width * Height * 3;
            for X := 0 to Width - 1 do
              for Y := 0 to Height - 1 do
              begin
                PFGetARGB(PF, PB^, A, R, G, B);
                C.R := R;
                C.G := G;
                C.B := B;
                Inc(PB);
                Inc(C);
              end;
            FreeMemNil(Bits);
            Bits := Data;
          end;
      end;
  end;
  FreeMemNil(Pal);
  Image.PalSize := 0;
  Image.ImageFormat := ifR8G8B8;
  Image.BitCount := 24;
end;

procedure Convert8To32(var Image: TImageData; DstFormat: TImageFormat);
var
  I, X, Y: Long;
  Data: Pointer;
  C: PARGBRec;
  PB: PByte;
  A, R, G, B: Byte;
  Pal: PPalArray;
  PF: TPixelFormat;

begin
  if not TestImage(Image) then
    Exit;

  case DstFormat of
    Low(if32bitARGB)..High(if32bitARGB):
      case Image.ImageFormat of
        ifIndex8,
          ifIntensity8:
          begin
            if Image.Palette = nil then
              Pal := GreyScalePal(256)
            else
              Pal := Image.Palette;
            for I := 0 to Image.LevelCount - 1 do
              with Image, Image.Levels[I] do
              begin
                GetMem(Data, Width * Height * 4);
                C := Data;
                Size := Width * Height * 4;
                PB := Bits;
                for X := 0 to Width - 1 do
                  for Y := 0 to Height - 1 do
                  begin
                    C.A := Pal[PB^].A;
                    C.R := Pal[PB^].R;
                    C.G := Pal[PB^].G;
                    C.B := Pal[PB^].B;
                    Inc(C);
                    Inc(PB);
                  end;
                FreeMemNil(Bits);
                Bits := Data;
              end;
          end;
        Low(if8bitARGBPF)..High(if8bitARGBPF):
          begin
            PF := ImageFormatToPF(Image.ImageFormat);
            for I := 0 to Image.LevelCount - 1 do
              with Image, Image.Levels[I] do
              begin
                PB := Bits;
                GetMem(Data, Width * Height * 4);
                C := Data;
                Size := Width * Height * 4;
                for X := 0 to Width - 1 do
                  for Y := 0 to Height - 1 do
                  begin
                    PFGetARGB(PF, PB^, A, R, G, B);
                    if PF.ABitCount > 0 then
                      C.A := A
                    else
                      C.A := $FF;
                    C.R := R;
                    C.G := G;
                    C.B := B;
                    Inc(PB);
                    Inc(C);
                  end;
                FreeMemNil(Bits);
                Bits := Data;
              end;
          end;
      end;
    ifIntensity32:
      begin
        for I := 0 to Image.LevelCount - 1 do
          with Image, Image.Levels[I] do
          begin
            PB := Bits;
            GetMem(Data, Width * Height * 4);
            C := Data;
            Size := Width * Height * 4;
            for X := 0 to Width - 1 do
              for Y := 0 to Height - 1 do
              begin
                C.Color := PB^ shl 24;
                Inc(PB);
                Inc(C);
              end;
            FreeMemNil(Bits);
            Bits := Data;
          end;
      end;
  end;
  FreeMemNil(Image.Palette);
  Image.Palette := nil;
  Image.PalSize := 0;
  Image.ImageFormat := DstFormat;
  Image.BitCount := 32;
end;

procedure Convert24To32(var Image: TImageData; DstFormat: TImageFormat);
var
  I, X, Y: Long;
  CA: PARGBRec;
  C: PRGBRec;
  Data: Pointer;
  A: Byte;

begin
  if not TestImage(Image) then
    Exit;

  A := Iff(ImageFormatToPF(DstFormat).ABitCount > 0, $FF, 0);
  for I := 0 to Image.LevelCount - 1 do
    with Image, Image.Levels[I] do
    begin
      GetMem(Data, Width * Height * 4);
      CA := Data;
      C := Bits;
      Size := Width * Height * 4;
      for X := 0 to Width - 1 do
        for Y := 0 to Height - 1 do
        begin
          CA.A := A;
          CA.R := C.R;
          CA.G := C.G;
          CA.B := C.B;
          Inc(CA);
          Inc(C);
        end;
      FreeMemNil(Bits);
      Bits := Data;
    end;
  Image.ImageFormat := DstFormat;
  Image.BitCount := 32;
end;

procedure Convert32To24(var Image: TImageData);
var
  I, X, Y: Long;
  CA: PARGBRec;
  C: PRGBRec;
  Data: Pointer;

begin
  if not TestImage(Image) then
    Exit;

  for I := 0 to Image.LevelCount - 1 do
    with Image, Image.Levels[I] do
    begin
      GetMem(Data, Width * Height * 3);
      C := Data;
      CA := Bits;
      Size := Width * Height * 3;
      for X := 0 to Width - 1 do
        for Y := 0 to Height - 1 do
        begin
          C.R := CA.R;
          C.G := CA.G;
          C.B := CA.B;
          Inc(CA);
          Inc(C);
        end;
      FreeMemNil(Bits);
      Bits := Data;
    end;
  Image.ImageFormat := ifR8G8B8;
  Image.BitCount := 24;
end;

procedure Convert8To16(var Image: TImageData; DstFormat: TImageFormat);
var
  I, X, Y: Long;
  Data: Pointer;
  W: PWord;
  B: PByte;
  C: TARGB;
  SrcPF, PF: TPixelFormat;
  Pal: PPalArray;
begin
  if not TestImage(Image) then
    Exit;

  Pal := nil;
  PF := ImageFormatToPF(DstFormat);
  case DstFormat of
    Low(if16bitARGBPF)..High(if16bitARGBPF):
      case Image.ImageFormat of
        ifIndex8, ifIntensity8:
          begin
            if Image.Palette = nil then
              Pal := GreyScalePal(256)
            else
              Pal := Image.Palette;

            for I := 0 to Image.LevelCount - 1 do
              with Image, Image.Levels[I] do
              begin
                GetMem(Data, Width * Height * 2);
                W := Data;
                Size := Width * Height * 2;
                B := Bits;
                for X := 0 to Width - 1 do
                  for Y := 0 to Height - 1 do
                  begin
                    W^ := PFSetARGB(PF, Pal[B^].Color);
                    Inc(W);
                    Inc(B);
                  end;
                FreeMemNil(Bits);
                Bits := Data;
              end;
          end;
        Low(if8bitARGBPF)..High(if8bitARGBPF):
          begin
            SrcPF := ImageFormatToPF(Image.ImageFormat);
            for I := 0 to Image.LevelCount - 1 do
              with Image, Image.Levels[I] do
              begin
                GetMem(Data, Width * Height * 2);
                W := Data;
                Size := Width * Height * 2;
                B := Bits;
                for X := 0 to Width - 1 do
                  for Y := 0 to Height - 1 do
                  begin
                    C := PFGetARGB(SrcPF, B^);
                    W^ := PFSetARGB(PF, C);
                    Inc(W);
                    Inc(B);
                  end;
                FreeMemNil(Bits);
                Bits := Data;
              end;
          end;
      end;
    ifIntensity16:
      for I := 0 to Image.LevelCount - 1 do
        with Image, Image.Levels[I] do
        begin
          GetMem(Data, Width * Height * 2);
          W := Data;
          Size := Width * Height * 2;
          B := Bits;
          for X := 0 to Width - 1 do
            for Y := 0 to Height - 1 do
            begin
              W^ := B^ shl 8;
              Inc(W);
              Inc(B);
            end;
          FreeMemNil(Bits);
          Bits := Data;
        end;
  end;
  FreeMemNil(Pal);
  Image.Palette := nil;
  Image.PalSize := 0;
  Image.ImageFormat := DstFormat;
  Image.BitCount := 16;
end;

procedure Convert24To16(var Image: TImageData; DstFormat: TImageFormat);
var
  I, X, Y: Long;
  C: PRGBRec;
  W: PWord;
  PF: TPixelFormat;
  Data: Pointer;

begin
  if not TestImage(Image) then
    Exit;

  PF := ImageFormatToPF(DstFormat);
  for I := 0 to Image.LevelCount - 1 do
    with Image, Image.Levels[I] do
    begin
      GetMem(Data, Width * Height * 2);
      W := Data;
      C := Bits;
      Size := Width * Height * 2;
      for X := 0 to Width - 1 do
        for Y := 0 to Height - 1 do
        begin
          W^ := PFSetARGB(PF, 0, C.R, C.G, C.B);
          Inc(C);
          Inc(W);
        end;
      FreeMemNil(Bits);
      Bits := Data;
    end;
  Image.ImageFormat := DstFormat;
  Image.BitCount := 16;
end;

procedure Convert32To16(var Image: TImageData; DstFormat: TImageFormat);
var
  I, X, Y: Long;
  CA: PARGBRec;
  W: PWord;
  PF: TPixelFormat;
  Data: Pointer;

begin
  if not TestImage(Image) then
    Exit;

  PF := ImageFormatToPF(DstFormat);
  for I := 0 to Image.LevelCount - 1 do
    with Image, Image.Levels[I] do
    begin
      GetMem(Data, Width * Height * 2);
      W := Data;
      CA := Bits;
      Size := Width * Height * 2;
      case DstFormat of
        Low(if16bitARGBPF)..High(if16bitARGBPF):
          for X := 0 to Width - 1 do
            for Y := 0 to Height - 1 do
            begin
              W^ := PFSetARGB(PF, CA.A, CA.R, CA.G, CA.B);
              Inc(CA);
              Inc(W);
            end;
        ifIntensity16:
          for X := 0 to Width - 1 do
            for Y := 0 to Height - 1 do
            begin
              W^ := CA.Color shr 16;
              Inc(CA);
              Inc(W);
            end;
      end;
      FreeMemNil(Bits);
      Bits := Data;
    end;
  Image.ImageFormat := DstFormat;
  Image.BitCount := 16;
end;

procedure Convert16To24(var Image: TImageData);
var
  I, X, Y: Long;
  C: PRGBRec;
  W: PWord;
  PF: TPixelFormat;
  Data: Pointer;
  A, R, G, B: Byte;
begin
  if not TestImage(Image) then
    Exit;

  PF := ImageFormatToPF(Image.ImageFormat);
  for I := 0 to Image.LevelCount - 1 do
    with Image, Image.Levels[I] do
    begin
      GetMem(Data, Width * Height * 3);
      C := Data;
      W := Bits;
      Size := Width * Height * 3;
      for X := 0 to Width - 1 do
        for Y := 0 to Height - 1 do
        begin
          PFGetARGB(PF, W^, A, R, G, B);
          C.R := R;
          C.G := G;
          C.B := B;
          Inc(C);
          Inc(W);
        end;
      FreeMemNil(Bits);
      Bits := Data;
    end;
  Image.ImageFormat := ifR8G8B8;
  Image.BitCount := 24;
end;

procedure Convert16To32(var Image: TImageData; DstFormat: TImageFormat);
var
  I, X, Y: Long;
  CA: PARGBRec;
  W: PWord;
  PF: TPixelFormat;
  Data: Pointer;
  A, R, G, B: Byte;

begin
  if not TestImage(Image) then
    Exit;

  PF := ImageFormatToPF(Image.ImageFormat);
  for I := 0 to Image.LevelCount - 1 do
    with Image, Image.Levels[I] do
    begin
      GetMem(Data, Width * Height * 4);
      CA := Data;
      W := Bits;
      Size := Width * Height * 4;
      case DstFormat of
        Low(if32bitARGB)..High(if32bitARGB):
          for X := 0 to Width - 1 do
            for Y := 0 to Height - 1 do
            begin
              PFGetARGB(PF, W^, A, R, G, B);
              CA.A := A;
              CA.R := R;
              CA.G := G;
              CA.B := B;
              Inc(CA);
              Inc(W);
            end;
        ifIntensity32:
          for X := 0 to Width - 1 do
            for Y := 0 to Height - 1 do
            begin
              CA.Color := W^ shl 16;
              Inc(CA);
              Inc(W);
            end;
      end;
      FreeMemNil(Bits);
      Bits := Data;
    end;
  Image.ImageFormat := DstFormat;
  Image.BitCount := 32;
end;

procedure Convert16To16(var Image: TImageData; DstFormat: TImageFormat);
var
  I, X, Y: Long;
  W: PWord;
  SrcPF, DstPF: TPixelFormat;
  C: TARGB;

begin
  if not TestImage(Image) then
    Exit;
  if Image.ImageFormat = DstFormat then
    Exit;

  SrcPF := ImageFormatToPF(Image.ImageFormat);
  DstPF := ImageFormatToPF(DstFormat);
  for I := 0 to Image.LevelCount - 1 do
    with Image, Image.Levels[I] do
    begin
      W := Bits;
      for X := 0 to Width - 1 do
        for Y := 0 to Height - 1 do
        begin
          C := PFGetARGB(SrcPF, W^);
          W^ := PFSetARGB(DstPF, C);
          Inc(W);
        end;
    end;
  Image.ImageFormat := DstFormat;
  Image.BitCount := 16;
end;

procedure Convert8To8(var Image: TImageData; DstFormat: TImageFormat);
var
  PF: TPixelFormat;
  I, X, Y: Long;
  B: PByte;

begin
  if not TestImage(Image) then
    Exit;
  if Image.ImageFormat = DstFormat then
    Exit;

  with Image do
    case ImageFormat of
      ifIntensity8:
        case DstFormat of
          ifIndex8:
            begin
              PalSize := 256;
              Palette := GreyScalePal(256, $FF);
            end;
          Low(if8bitARGBPF)..High(if8bitARGBPF):
            begin
              PF := ImageFormatToPF(DstFormat);
              for I := 0 to LevelCount - 1 do
                with Levels[I] do
                begin
                  B := Bits;
                  for X := 0 to Width - 1 do
                    for Y := 0 to Height - 1 do
                    begin
                      B^ := PFSetARGB(PF, ARGB($FF, B^, B^, B^));
                      Inc(B);
                    end;
                end;
            end;
        end;
      ifIndex8:
        case DstFormat of
          ifIntensity8: ConvertToGreyScale8(Image);
          Low(if8bitARGBPF)..High(if8bitARGBPF):
            begin
              PF := ImageFormatToPF(DstFormat);
              for I := 0 to LevelCount - 1 do
                with Levels[I] do
                begin
                  B := Bits;
                  for X := 0 to Width - 1 do
                    for Y := 0 to Height - 1 do
                    begin
                      B^ := PFSetARGB(PF, Palette[B^].Color);
                      Inc(B);
                    end;
                end;
              PalSize := 0;
              FreeMemNil(Palette);
            end;
        end;
      Low(if8bitARGBPF)..High(if8bitARGBPF):
        case DstFormat of
          ifIntensity8: ConvertToGreyScale8(Image);
          ifIndex8:
            begin
              Palette := HalftonePal(256, 3, 3, 2, $FF);
              PalSize := 256;
              PF := ImageFormatToPF(ImageFormat);
              for I := 0 to LevelCount - 1 do
                with Levels[I] do
                begin
                  B := Bits;
                  for X := 0 to Width - 1 do
                    for Y := 0 to Height - 1 do
                    begin
                      B^ := ClosestColorInPal(Palette, 256, PFGetARGB(PF, B^));
                      Inc(B);
                    end;
                end;
            end;
        end;
    end;
  Image.ImageFormat := DstFormat;
end;

procedure Convert16To8(var Image: TImageData; DstFormat: TImageFormat);
var
  Data: Pointer;
  I, X, Y: Long;
  B: PByte;
  W: PWord;
  SrcPF, DstPF: TPixelFormat;

begin
  case DstFormat of
    ifIntensity8:
      if Image.ImageFormat <> ifIntensity16 then
        ConvertToGreyScale8(Image)
      else
      begin
        for I := 0 to Image.LevelCount - 1 do
          with Image, Image.Levels[I] do
          begin
            GetMem(Data, Width * Height);
            W := Bits;
            Size := Width * Height;
            B := Data;
            for X := 0 to Width - 1 do
              for Y := 0 to Height - 1 do
              begin
                B^ := W^ shr 8;
                Inc(B);
                Inc(W);
              end;
            FreeMemNil(Bits);
            Bits := Data;
          end;
      end;
    ifIndex8: ConvertARGBToIndex(Image);
    Low(if8bitARGBPF)..High(if8bitARGBPF):
      begin
        SrcPF := ImageFormatToPF(Image.ImageFormat);
        DstPF := ImageFormatToPF(DstFormat);
        for I := 0 to Image.LevelCount - 1 do
          with Image, Image.Levels[I] do
          begin
            GetMem(Data, Width * Height);
            W := Bits;
            Size := Width * Height;
            B := Data;
            for X := 0 to Width - 1 do
              for Y := 0 to Height - 1 do
              begin
                B^ := PFSetARGB(DstPF, PFGetARGB(SrcPF, W^));
                Inc(B);
                Inc(W);
              end;
            FreeMemNil(Bits);
            Bits := Data;
          end;
      end;
  end;
  Image.ImageFormat := DstFormat;
  Image.BitCount := 8;
end;

procedure Convert24To8(var Image: TImageData; DstFormat: TImageFormat);
var
  Data: Pointer;
  I, X, Y: Long;
  B: PByte;
  C: PRGBRec;
  PF: TPixelFormat;

begin
  case DstFormat of
    ifIntensity8: ConvertToGreyScale8(Image);
    ifIndex8: ConvertARGBToIndex(Image);
    Low(if8bitARGBPF)..High(if8bitARGBPF):
      begin
        PF := ImageFormatToPF(DstFormat);
        for I := 0 to Image.LevelCount - 1 do
          with Image, Image.Levels[I] do
          begin
            GetMem(Data, Width * Height);
            C := Bits;
            Size := Width * Height;
            B := Data;
            for X := 0 to Width - 1 do
              for Y := 0 to Height - 1 do
              begin
                B^ := PFSetARGB(PF, $FF, C.R, C.G, C.B);
                Inc(B);
                Inc(C);
              end;
            FreeMemNil(Bits);
            Bits := Data;
          end;
      end;
  end;
  Image.ImageFormat := DstFormat;
  Image.BitCount := 8;
end;

procedure Convert32To8(var Image: TImageData; DstFormat: TImageFormat);
var
  Data: Pointer;
  I, X, Y: Long;
  B: PByte;
  C: PARGBRec;
  PF: TPixelFormat;

begin
  case DstFormat of
    ifIntensity8:
      if Image.ImageFormat <> ifIntensity32 then
        ConvertToGreyScale8(Image)
      else
      begin
        for I := 0 to Image.LevelCount - 1 do
          with Image, Image.Levels[I] do
          begin
            GetMem(Data, Width * Height);
            C := Bits;
            Size := Width * Height;
            B := Data;
            for X := 0 to Width - 1 do
              for Y := 0 to Height - 1 do
              begin
                B^ := C.Color shr 24;
                Inc(B);
                Inc(C);
              end;
            FreeMemNil(Bits);
            Bits := Data;
          end;
      end;
    ifIndex8: ConvertARGBToIndex(Image);
    Low(if8bitARGBPF)..High(if8bitARGBPF):
      begin
        PF := ImageFormatToPF(DstFormat);
        for I := 0 to Image.LevelCount - 1 do
          with Image, Image.Levels[I] do
          begin
            GetMem(Data, Width * Height);
            C := Bits;
            Size := Width * Height;
            B := Data;
            for X := 0 to Width - 1 do
              for Y := 0 to Height - 1 do
              begin
                B^ := PFSetARGB(PF, C.A, C.R, C.G, C.B);
                Inc(B);
                Inc(C);
              end;
            FreeMemNil(Bits);
            Bits := Data;
          end;
      end;
  end;
  Image.ImageFormat := DstFormat;
  Image.BitCount := 8;
end;

procedure ConvertToGreyScale8(var Image: TImageData);
var
  RCh, GCh, BCh: array[Byte] of Byte;
  I, X, Y: Long;
  W: PWord;
  C: PRGBRec;
  CA: PARGBRec;
  PF: TPixelFormat;
  Data: Pointer;
  PB, PBS: PByte;
  A, R, G, B: Byte;

begin
  if not TestImage(Image) then
    Exit;

  if Image.ImageFormat = ifIntensity8 then
    Exit;

  for I := 0 to 255 do
  begin
    RCh[I] := Trunc(0.3588 * I);
    GCh[I] := Trunc(0.4020 * I);
    BCh[I] := Trunc(0.2392 * I);
  end;

  for I := 0 to Image.LevelCount - 1 do
    with Image, Image.Levels[I] do
    begin
      GetMem(Data, Width * Height);
      PB := Data;
      Size := Width * Height;
      case BitCount of
        8: case ImageFormat of
            ifIndex8:
              begin
                PBS := Bits;
                for X := 0 to Width - 1 do
                  for Y := 0 to Height - 1 do
                  begin
                    PB^ := RCh[Palette[PBS^].R] + GCh[Palette[PBS^].G] +
                      BCh[Palette[PBS^].B];
                    Inc(PB);
                    Inc(PBS);
                  end;
              end;
            Low(if8bitARGBPF)..High(if8bitARGBPF):
              begin
                PF := ImageFormatToPF(ImageFormat);
                PBS := Bits;
                for X := 0 to Width - 1 do
                  for Y := 0 to Height - 1 do
                  begin
                    PFGetARGB(PF, PBS^, A, R, G, B);
                    PB^ := RCh[R] + GCh[G] + BCh[B];
                    Inc(PBS);
                    Inc(PB);
                  end;
              end;
          end;
        16:
          begin
            PF := ImageFormatToPF(ImageFormat);
            W := Bits;
            for X := 0 to Width - 1 do
              for Y := 0 to Height - 1 do
              begin
                PFGetARGB(PF, W^, A, R, G, B);
                PB^ := RCh[R] + GCh[G] + BCh[B];
                Inc(W);
                Inc(PB);
              end;
          end;
        24:
          begin
            C := Bits;
            for X := 0 to Width - 1 do
              for Y := 0 to Height - 1 do
              begin
                PB^ := RCh[C.R] + GCh[C.G] + BCh[C.B];
                Inc(PB);
                Inc(C);
              end;
          end;
        32:
          begin
            CA := Bits;
            for X := 0 to Width - 1 do
              for Y := 0 to Height - 1 do
              begin
                PB^ := RCh[CA.R] + GCh[CA.G] + BCh[CA.B];
                Inc(PB);
                Inc(CA);
              end;
          end;
      end;
      FreeMemNil(Bits);
      Bits := Data;
    end;

  FreeMemNil(Image.Palette);
  Image.PalSize := 0;
  Image.BitCount := 8;
  Image.ImageFormat := ifIntensity8;
end;

procedure ConvertARGBToIndex(var Image: TImageData);
var
  Data: Pointer;
  I, X, Y: Long;
  B, BS: PByte;
  C: TARGB;

  function GetColor: TARGB;
  var
    PF: TPixelFormat;
  begin
    Result := 0;
    with Image, Image.Levels[I] do
      case BitCount of
        8: case ImageFormat of
            ifIndex8: Result := Palette[BS^].Color;
            ifIntensity8: Result := ($FF shl 24) or (BS^ shl 16) or (BS^ shl 8)
              or BS^;
            Low(if8bitARGBPF)..High(if8bitARGBPF):
              begin
                PF := ImageFormatToPF(ImageFormat);
                Result := PFGetARGB(PF, BS^);
              end;
          end;
        16:
          begin
            PF := ImageFormatToPF(ImageFormat);
            Result := PFGetARGB(PF, PWord(Pointer(BS))^);
          end;
        24: with PRGBRec(Pointer(BS))^ do
            Result := (R shl 16) or (G shl 8) or B;
        32: Result := PULong(Pointer(BS))^;
      end;
  end;

begin
  Image.PalSize := 256;
  FreeMemNil(Image.Palette);
  Image.Palette := HalftonePal(256, 2, 4, 2, $FF);

  for I := 0 to Image.LevelCount - 1 do
    with Image, Image.Levels[I] do
    begin
      GetMem(Data, Width * Height);
      Size := Width * Height;
      B := Data;
      BS := Bits;
      for X := 0 to Width - 1 do
        for Y := 0 to Height - 1 do
        begin
          C := GetColor;
          B^ := ClosestColorInPal(Palette, 256, C);
          Inc(B);
          Inc(BS, BitCount shr 3);
        end;
      FreeMemNil(Bits);
      Bits := Data;
    end;
  Image.BitCount := 8;
  Image.ImageFormat := ifIndex8;
end;

procedure Convert16To16Custom(var Image: TImageData; const DstPF: TPixelFormat);
var
  I, X, Y: Long;
  W: PWord;
  SrcPF: TPixelFormat;
  C: TARGB;

begin
  if not TestImage(Image) then
    Exit;

  SrcPF := ImageFormatToPF(Image.ImageFormat);
  for I := 0 to Image.LevelCount - 1 do
    with Image, Image.Levels[I] do
    begin
      W := Bits;
      for X := 0 to Width - 1 do
        for Y := 0 to Height - 1 do
        begin
          C := PFGetARGB(SrcPF, W^);
          W^ := PFSetARGB(DstPF, C);
          Inc(W);
        end;
    end;
  Image.ImageFormat := ifCustom;
  Image.BitCount := 16;
end;

procedure Convert16CustomTo16(var Image: TImageData; const SrcPF: TPixelFormat;
  DstFormat: TImageFormat);
var
  I, X, Y: Long;
  W: PWord;
  DstPF: TPixelFormat;
  C: TARGB;

begin
  if not TestImage(Image) then
    Exit;

  DstPF := ImageFormatToPF(DstFormat);
  for I := 0 to Image.LevelCount - 1 do
    with Image, Image.Levels[I] do
    begin
      W := Bits;
      for X := 0 to Width - 1 do
        for Y := 0 to Height - 1 do
        begin
          C := PFGetARGB(SrcPF, W^);
          W^ := PFSetARGB(DstPF, C);
          Inc(W);
        end;
    end;
  Image.ImageFormat := DstFormat;
  Image.BitCount := 16;
end;

procedure ResizePixels(NewWidth, NewHeight: Long; var Image: TImageData;
  Level: Long;
  UseScaleRatio: boolean);
var
  X, Y, Xp, Yp, Sx, Sy: Long;
  BytesPP: Byte;
  Pc: PULong;
  DestW, DestH, BytesPL: Long;
  Mi, Ma, I: Long;
  Data, Line: Pointer;
  ScaleX, ScaleY: single;
begin
  Mi := Iff(Level <> $FFFF, Level, 0);
  Ma := Iff(Level <> $FFFF, Level, Image.LevelCount - 1);
  ScaleX := NewWidth / Image.Levels[0].Width;
  ScaleY := NewHeight / Image.Levels[0].Height;
  BytesPP := Image.BitCount shr 3;
  for I := Mi to Ma do
  begin
    DestW := Iff(UseScaleRatio, Round(Image.Levels[I].Width * ScaleX),
      NewWidth);
    DestH := Iff(UseScaleRatio, Round(Image.Levels[I].Height * ScaleY),
      NewHeight);
    BytesPL := Image.Levels[I].Width * BytesPP;
    GetMem(Data, DestW * DestH * BytesPP);
    Sx := (Image.Levels[I].Width shl 16) div DestW;
    Sy := (Image.Levels[I].Height shl 16) div DestH;
    Yp := 0;
    Pc := Data;
    for Y := 0 to DestH - 1 do
    begin
      Xp := 0;
      Line := Pointer(Long(Image.Levels[I].Bits) + (Yp shr 16) * BytesPL);
      for X := 0 to DestW - 1 do
      begin
        case BytesPP of
          1: PByte(Pc)^ := PByte(Long(Line) + (Xp shr 16) * BytesPP)^;
          2: PWord(Pc)^ := PWord(Long(Line) + (Xp shr 16) * BytesPP)^;
          3: PRGBRec(Pc)^ := PRGBRec(Long(Line) + (Xp shr 16) * BytesPP)^;
          4: Pc^ := PULong(Long(Line) + (Xp shr 16) * BytesPP)^;
        end;
        Pc := Pointer(Long(Pc) + BytesPP);
        Inc(Xp, Sx);
      end;
      Inc(Yp, Sy);
    end;
    Image.Levels[I].Width := DestW;
    Image.Levels[I].Height := DestH;
    Image.Levels[I].Size := DestW * DestH * BytesPP;
    FreeMem(Image.Levels[I].Bits);
    Image.Levels[I].Bits := Data;
  end;
end;

procedure ResizeBilinear(NewWidth, NewHeight: Long; var Image: TImageData;
  Level: Long = $FFFF; UseScaleRatio: boolean = false);
var
  Z, Z2, Iz2, W1, W2, W3, W4: Long;
  X, Y, Xp, Yp, Xpp, Ypp, T, TT: Long;
  BytesPP: Byte;
  Pc: PULong;
  DestW, DestH, BytesPL: Long;
  Mi, Ma, I: Long;
  Data, Line1, Line2: Pointer;
  ScaleX, ScaleY: single;
  C1, C2, C3, C4: TARGB;
  A, R, G, B: Byte;
  PF: TPixelFormat;

begin

  ScaleX := NewWidth / Image.Levels[0].Width;
  ScaleY := NewHeight / Image.Levels[0].Height;

  if not ((ScaleX > 1.0) or (ScaleY > 1.0)) then
  begin
    ResizePixels(NewWidth, NewHeight, Image, Level, UseScaleRatio);
    FilterBilinear(Image, Level);
    Exit;
  end;

  BytesPP := Image.BitCount shr 3;
  PF := ImageFormatToPF(Image.ImageFormat);

  Mi := Iff(Level <> $FFFF, Level, 0);
  Ma := Iff(Level <> $FFFF, Level, Image.LevelCount - 1);

  for I := Mi to Ma do
  begin
    DestW := Iff(UseScaleRatio, Round(Image.Levels[I].Width * ScaleX),
      NewWidth);
    DestH := Iff(UseScaleRatio, Round(Image.Levels[I].Height * ScaleY),
      NewHeight);
    BytesPL := Image.Levels[I].Width * BytesPP;
    GetMem(Data, DestW * DestH * BytesPP);
    Xpp := (Image.Levels[I].Width shl 16) div DestW;
    Ypp := (Image.Levels[I].Height shl 16) div DestH;
    Yp := 0;
    Pc := Data;

    for Y := 0 to DestH - 1 do
    begin
      Xp := Yp shr 16;
      Line1 := Pointer(Long(Image.Levels[I].Bits) + Xp * BytesPL);
      if Xp < Image.Levels[I].Height - 1 then
        Line2 := Pointer(Long(Image.Levels[I].Bits) + (Xp + 1) * BytesPL)
      else
        // to keep seamless images seamless
        Line2 := Image.Levels[I].Bits;

      Xp := 0;
      Z2 := (Yp and $FFFF) + 1;
      Iz2 := ((not Yp) and $FFFF) + 1;
      for X := 0 to DestW - 1 do
      begin
        T := Xp shr 16;
        TT := T + 1;

        // to avoid odd color of last pixel
        if (TT = Image.Levels[I].Width) then
          TT := 0;

        Z := Xp and $FFFF;
        W2 := (Iz2 * Z) shr 16;
        W1 := Iz2 - W2;
        W4 := (Z2 * Z) shr 16;
        W3 := Z2 - W4;
        case BytesPP of
          1: case Image.ImageFormat of
              ifIntensity8:
                PByte(Pc)^ := (PByteArray(Line1)[T] * W1 +
                  PByteArray(Line1)[TT] * W2 +
                  PByteArray(Line2)[T] * W3 + PByteArray(Line2)[TT] * W4) shr
                  16;
              Low(if8bitARGBPF)..High(if8bitARGBPF):
                begin
                  C1 := PFGetARGB(PF, PByteArray(Line1)[T]);
                  C2 := PFGetARGB(PF, PByteArray(Line1)[TT]);
                  C3 := PFGetARGB(PF, PByteArray(Line2)[T]);
                  C4 := PFGetARGB(PF, PByteArray(Line2)[TT]);

                  A := (Byte(C1 shr 24) + Byte(C2 shr 24) + Byte(C3 shr 24) +
                    Byte(C4 shr 24)) div 4;
                  R := (Byte(C1 shr 16) + Byte(C2 shr 16) + Byte(C3 shr 16) +
                    Byte(C4 shr 16)) div 4;
                  G := (Byte(C1 shr 8) + Byte(C2 shr 8) + Byte(C3 shr 8) +
                    Byte(C4 shr 8)) div 4;
                  B := (Byte(C1) + Byte(C2) + Byte(C3) + Byte(C4)) div 4;

                  PByte(Pc)^ := PFSetARGB(PF, A, R, G, B);
                end;
              ifIndex8:
                begin
                  A := (Image.Palette[PByteArray(Line1)[T]].A +
                    Image.Palette[PByteArray(Line1)[TT]].A +
                    Image.Palette[PByteArray(Line2)[T]].A +
                    Image.Palette[PByteArray(Line2)[TT]].A) div 4;
                  R := (Image.Palette[PByteArray(Line1)[T]].R +
                    Image.Palette[PByteArray(Line1)[TT]].R +
                    Image.Palette[PByteArray(Line2)[T]].R +
                    Image.Palette[PByteArray(Line2)[TT]].R) div 4;
                  G := (Image.Palette[PByteArray(Line1)[T]].G +
                    Image.Palette[PByteArray(Line1)[TT]].G +
                    Image.Palette[PByteArray(Line2)[T]].G +
                    Image.Palette[PByteArray(Line2)[TT]].G) div 4;
                  B := (Image.Palette[PByteArray(Line1)[T]].B +
                    Image.Palette[PByteArray(Line1)[TT]].B +
                    Image.Palette[PByteArray(Line2)[T]].B +
                    Image.Palette[PByteArray(Line2)[TT]].B) div 4;

                  PByte(Pc)^ := ClosestColorInPal(Image.Palette, Image.PalSize,
                    ARGB(A, R, G, B));
                end;
            end;
          2: case Image.ImageFormat of
              ifIntensity16:
                begin
                  PWord(Pc)^ := (PWordArray(Line1)[T] + PWordArray(Line1)[TT] +
                    PWordArray(Line2)[T] + PWordArray(Line2)[TT]) div 4;
                end;
              Low(if16bitARGBPF)..High(if16bitARGBPF):
                begin
                  C1 := PFGetARGB(PF, PWordArray(Line1)[T]);
                  C2 := PFGetARGB(PF, PWordArray(Line1)[TT]);
                  C3 := PFGetARGB(PF, PWordArray(Line2)[T]);
                  C4 := PFGetARGB(PF, PWordArray(Line2)[TT]);

                  A := (Byte(C1 shr 24) + Byte(C2 shr 24) + Byte(C3 shr 24) +
                    Byte(C4
                    shr 24)) div 4;
                  R := (Byte(C1 shr 16) + Byte(C2 shr 16) + Byte(C3 shr 16) +
                    Byte(C4
                    shr 16)) div 4;
                  G := (Byte(C1 shr 8) + Byte(C2 shr 8) + Byte(C3 shr 8) +
                    Byte(C4
                    shr 8)) div 4;
                  B := (Byte(C1) + Byte(C2) + Byte(C3) + Byte(C4)) div 4;

                  PWord(Pc)^ := PFSetARGB(PF, A, R, G, B);
                end;
            end;
          3:
            begin
              PRGBRec(Pc).R := (PRGBRecArray(Line1)[T].R * W1 +
                PRGBRecArray(Line1)[TT].R * W2 +
                PRGBRecArray(Line2)[T].R * W3 + PRGBRecArray(Line2)[TT].R *
                W4) shr 16;
              PRGBRec(Pc).G := (PRGBRecArray(Line1)[T].G * W1 +
                PRGBRecArray(Line1)[TT].G * W2 +
                PRGBRecArray(Line2)[T].G * W3 + PRGBRecArray(Line2)[TT].G *
                W4) shr 16;
              PRGBRec(Pc).B := (PRGBRecArray(Line1)[T].B * W1 +
                PRGBRecArray(Line1)[TT].B * W2 +
                PRGBRecArray(Line2)[T].B * W3 + PRGBRecArray(Line2)[TT].B *
                W4) shr 16;
            end;
          4: case Image.ImageFormat of
              Low(if32bitARGB)..High(if32bitARGB):
                begin
                  PARGBRec(Pc).A := (PARGBRecArray(Line1)[T].A * W1 +
                    PARGBRecArray(Line1)[TT].A * W2 +
                    PARGBRecArray(Line2)[T].A * W3 + PARGBRecArray(Line2)[TT].A
                    * W4) shr 16;
                  PARGBRec(Pc).R := (PARGBRecArray(Line1)[T].R * W1 +
                    PARGBRecArray(Line1)[TT].R * W2 +
                    PARGBRecArray(Line2)[T].R * W3 + PARGBRecArray(Line2)[TT].R
                    * W4) shr 16;
                  PARGBRec(Pc).G := (PARGBRecArray(Line1)[T].G * W1 +
                    PARGBRecArray(Line1)[TT].G * W2 +
                    PARGBRecArray(Line2)[T].G * W3 + PARGBRecArray(Line2)[TT].G
                    * W4) shr 16;
                  PARGBRec(Pc).B := (PARGBRecArray(Line1)[T].B * W1 +
                    PARGBRecArray(Line1)[TT].B * W2 +
                    PARGBRecArray(Line2)[T].B * W3 + PARGBRecArray(Line2)[TT].B
                    * W4) shr 16;
                end;
              ifIntensity32:
                begin
                  Pc^ := (Int64(PULongArray(Line1)[T]) * W1 +
                    Int64(PULongArray(Line1)[TT]) * W2 +
                    Int64(PULongArray(Line2)[T]) * W3 +
                    Int64(PULongArray(Line2)[TT]) * W4) shr 16;
                end;
            end;
        end;
        Inc(Xp, Xpp);
        Pc := Pointer(Long(Pc) + BytesPP);
      end;
      Inc(Yp, Ypp);
    end;
    Image.Levels[I].Width := DestW;
    Image.Levels[I].Height := DestH;
    Image.Levels[I].Size := DestW * DestH * BytesPP;
    FreeMem(Image.Levels[I].Bits);
    Image.Levels[I].Bits := Data;
  end;
end;

procedure FilterBilinear(var Image: TImageData; Level: Long);
var
  Mi, Ma, I: Long;
  Data: Pointer;
  Line1, Line2: Pointer;
  XX, X, Y, W, H, BytesPP: Long;
  Pc: PULong;
  PF: TPixelFormat;
  C1, C2, C3, C4: TARGB;
  A, R, G, B: Byte;

begin
  Mi := Iff(Level <> $FFFF, Level, 0);
  Ma := Iff(Level <> $FFFF, Level, Image.LevelCount - 1);

  BytesPP := Image.BitCount shr 3;

  PF := ImageFormatToPF(Image.ImageFormat);

  for I := Mi to Ma do
  begin
    W := Image.Levels[I].Width;
    H := Image.Levels[I].Height;
    GetMem(Data, Image.Levels[I].Size);
    Pc := Data;

    for Y := 0 to H - 1 do
    begin
      Line1 := Pointer(Long(Image.Levels[I].Bits) + Y * W * BytesPP);
      if Y <> H - 1 then
        Line2 := Pointer(Long(Image.Levels[I].Bits) + (Y + 1) * W * BytesPP)
      else
        // to keep seamless images seamless
        Line2 := Image.Levels[I].Bits;

      for X := 0 to W - 1 do
      begin
        XX := X + 1;
        // to avoid odd color of last pixel
        if (X = W - 1) then
          XX := 0;

        case BytesPP of
          1: case Image.ImageFormat of
              ifIntensity8:
                begin
                  PByte(Pc)^ := (PByteArray(Line1)[X] + PByteArray(Line1)[XX] +
                    PByteArray(Line2)[X] + PByteArray(Line2)[XX]) div 4;
                end;
              Low(if8bitARGBPF)..High(if8bitARGBPF):
                begin
                  C1 := PFGetARGB(PF, PByteArray(Line1)[X]);
                  C2 := PFGetARGB(PF, PByteArray(Line1)[XX]);
                  C3 := PFGetARGB(PF, PByteArray(Line2)[X]);
                  C4 := PFGetARGB(PF, PByteArray(Line2)[XX]);

                  A := (Byte(C1 shr 24) + Byte(C2 shr 24) + Byte(C3 shr 24) +
                    Byte(C4 shr 24)) div 4;
                  R := (Byte(C1 shr 16) + Byte(C2 shr 16) + Byte(C3 shr 16) +
                    Byte(C4 shr 16)) div 4;
                  G := (Byte(C1 shr 8) + Byte(C2 shr 8) + Byte(C3 shr 8) +
                    Byte(C4 shr 8)) div 4;
                  B := (Byte(C1) + Byte(C2) + Byte(C3) + Byte(C4)) div 4;

                  PByte(Pc)^ := PFSetARGB(PF, A, R, G, B);
                end;
              ifIndex8:
                begin
                  A := (Image.Palette[PByteArray(Line1)[X]].A +
                    Image.Palette[PByteArray(Line1)[XX]].A +
                    Image.Palette[PByteArray(Line2)[X]].A +
                    Image.Palette[PByteArray(Line2)[XX]].A) div 4;
                  R := (Image.Palette[PByteArray(Line1)[X]].R +
                    Image.Palette[PByteArray(Line1)[XX]].R +
                    Image.Palette[PByteArray(Line2)[X]].R +
                    Image.Palette[PByteArray(Line2)[XX]].R) div 4;
                  G := (Image.Palette[PByteArray(Line1)[X]].G +
                    Image.Palette[PByteArray(Line1)[XX]].G +
                    Image.Palette[PByteArray(Line2)[X]].G +
                    Image.Palette[PByteArray(Line2)[XX]].G) div 4;
                  B := (Image.Palette[PByteArray(Line1)[X]].B +
                    Image.Palette[PByteArray(Line1)[XX]].B +
                    Image.Palette[PByteArray(Line2)[X]].B +
                    Image.Palette[PByteArray(Line2)[XX]].B) div 4;

                  PByte(Pc)^ := ClosestColorInPal(Image.Palette, Image.PalSize,
                    ARGB(A, R, G, B));
                end;
            end;
          2: case Image.ImageFormat of
              ifIntensity16:
                begin
                  PWord(Pc)^ := (PWordArray(Line1)[X] + PWordArray(Line1)[XX] +
                    PWordArray(Line2)[X] + PWordArray(Line2)[XX]) div 4;
                end;
              Low(if16bitARGBPF)..High(if16bitARGBPF):
                begin
                  C1 := PFGetARGB(PF, PWordArray(Line1)[X]);
                  C2 := PFGetARGB(PF, PWordArray(Line1)[XX]);
                  C3 := PFGetARGB(PF, PWordArray(Line2)[X]);
                  C4 := PFGetARGB(PF, PWordArray(Line2)[XX]);

                  A := (Byte(C1 shr 24) + Byte(C2 shr 24) + Byte(C3 shr 24) +
                    Byte(C4
                    shr 24)) div 4;
                  R := (Byte(C1 shr 16) + Byte(C2 shr 16) + Byte(C3 shr 16) +
                    Byte(C4
                    shr 16)) div 4;
                  G := (Byte(C1 shr 8) + Byte(C2 shr 8) + Byte(C3 shr 8) +
                    Byte(C4
                    shr 8)) div 4;
                  B := (Byte(C1) + Byte(C2) + Byte(C3) + Byte(C4)) div 4;

                  PWord(Pc)^ := PFSetARGB(PF, A, R, G, B);
                end;
            end;
          3:
            begin
              PRGBRec(Pc).R := (PRGBRecArray(Line1)[X].R +
                PRGBRecArray(Line1)[XX].R +
                PRGBRecArray(Line2)[X].R + PRGBRecArray(Line2)[XX].R) div 4;
              PRGBRec(Pc).G := (PRGBRecArray(Line1)[X].G +
                PRGBRecArray(Line1)[XX].G +
                PRGBRecArray(Line2)[X].G + PRGBRecArray(Line2)[XX].G) div 4;
              PRGBRec(Pc).B := (PRGBRecArray(Line1)[X].B +
                PRGBRecArray(Line1)[XX].B +
                PRGBRecArray(Line2)[X].B + PRGBRecArray(Line2)[XX].B) div 4;
            end;
          4: case Image.ImageFormat of
              Low(if32bitARGB)..High(if32bitARGB):
                begin
                  PARGBRec(Pc).A := (PARGBRecArray(Line1)[X].A +
                    PARGBRecArray(Line1)[XX].A +
                    PARGBRecArray(Line2)[X].A + PARGBRecArray(Line2)[XX].A)
                    div 4;
                  PARGBRec(Pc).R := (PARGBRecArray(Line1)[X].R +
                    PARGBRecArray(Line1)[XX].R +
                    PARGBRecArray(Line2)[X].R + PARGBRecArray(Line2)[XX].R)
                    div 4;
                  PARGBRec(Pc).G := (PARGBRecArray(Line1)[X].G +
                    PARGBRecArray(Line1)[XX].G +
                    PARGBRecArray(Line2)[X].G + PARGBRecArray(Line2)[XX].G)
                    div 4;
                  PARGBRec(Pc).B := (PARGBRecArray(Line1)[X].B +
                    PARGBRecArray(Line1)[XX].B +
                    PARGBRecArray(Line2)[X].B + PARGBRecArray(Line2)[XX].B)
                    div 4;
                end;
              ifIntensity32:
                begin
                  Pc^ := (Int64(PULongArray(Line1)[X]) +
                    Int64(PULongArray(Line1)[XX]) +
                    Int64(PULongArray(Line2)[X]) + Int64(PULongArray(Line2)[XX]))
                    div 4;
                end;
            end;
        end;
        Pc := Pointer(Long(Pc) + BytesPP);
      end;
    end;

    FreeMem(Image.Levels[I].Bits);
    Image.Levels[I].Bits := Data;
  end;
end;

procedure FilterTrilinear(var Image: TImageData; Level: Long = $FFFF);
var
  Mi, Ma, I: Long;
  Data: Pointer;
  LineM1, Line, LineP1: Pointer;
  XM1, X, XP1, Y, W, H, BytesPP: Long;
  Pc: PULong;
  PF: TPixelFormat;
  C1, C2, C3, C4, C5, C6, C7, C8, C9: TARGB;
  A, R, G, B: Byte;

begin
  Mi := Iff(Level <> $FFFF, Level, 0);
  Ma := Iff(Level <> $FFFF, Level, Image.LevelCount - 1);

  BytesPP := Image.BitCount shr 3;

  PF := ImageFormatToPF(Image.ImageFormat);

  for I := Mi to Ma do
  begin
    W := Image.Levels[I].Width;
    H := Image.Levels[I].Height;
    GetMem(Data, Image.Levels[I].Size);
    Pc := Data;

    for Y := 0 to H - 1 do
    begin
      if Y <> 0 then
        LineM1 := Pointer(Long(Image.Levels[I].Bits) + (Y - 1) * W * BytesPP)
      else
       // to keep seamless images seamless
        LineM1 := Pointer(Long(Image.Levels[I].Bits) + (H - 1) * W * BytesPP);
      Line := Pointer(Long(Image.Levels[I].Bits) + Y * W * BytesPP);
      if Y <> H - 1 then
        LineP1 := Pointer(Long(Image.Levels[I].Bits) + (Y + 1) * W * BytesPP)
      else
        // to keep seamless images seamless
        LineP1 := Image.Levels[I].Bits;

      for X := 0 to W - 1 do
      begin
        XM1 := X - 1;
        XP1 := X + 1;

        // to avoid odd color of first/last pixel
        if (X = W - 1) then
          XP1 := 0;
        if (X = 0) then
          XM1 := W - 1;

        case BytesPP of
          1: case Image.ImageFormat of
              ifIntensity8:
                begin
                  PByte(Pc)^ := (PByteArray(LineM1)[XM1] +
                    PByteArray(LineM1)[X] + PByteArray(LineM1)[XP1] +
                    PByteArray(Line)[XM1] +
                    PByteArray(Line)[X] + PByteArray(Line)[XP1] +
                    PByteArray(LineP1)[XM1] +
                    PByteArray(LineP1)[X] + PByteArray(LineP1)[XP1]) div 9;
                end;
              Low(if8bitARGBPF)..High(if8bitARGBPF):
                begin
                  C1 := PFGetARGB(PF, PByteArray(LineM1)[XM1]);
                  C2 := PFGetARGB(PF, PByteArray(LineM1)[X]);
                  C3 := PFGetARGB(PF, PByteArray(LineM1)[XP1]);
                  C4 := PFGetARGB(PF, PByteArray(Line)[XM1]);
                  C5 := PFGetARGB(PF, PByteArray(Line)[X]);
                  C6 := PFGetARGB(PF, PByteArray(Line)[XP1]);
                  C7 := PFGetARGB(PF, PByteArray(LineP1)[XM1]);
                  C8 := PFGetARGB(PF, PByteArray(LineP1)[X]);
                  C9 := PFGetARGB(PF, PByteArray(LineP1)[XP1]);

                  A := (Byte(C1 shr 24) + Byte(C2 shr 24) + Byte(C3 shr 24) +
                    Byte(C4 shr 24) + Byte(C5 shr 24) + Byte(C6 shr 24)
                    + Byte(C7 shr 24) + Byte(C8 shr 24) + Byte(C9 shr 24)) div
                    9;
                  R := (Byte(C1 shr 16) + Byte(C2 shr 16) + Byte(C3 shr 16) +
                    Byte(C4 shr 16) + Byte(C5 shr 16) + Byte(C6 shr 16)
                    + Byte(C7 shr 16) + Byte(C8 shr 16) + Byte(C9 shr 16)) div
                    9;
                  G := (Byte(C1 shr 8) + Byte(C2 shr 8) + Byte(C3 shr 8) +
                    Byte(C4 shr 8) + Byte(C5 shr 8) + Byte(C6 shr 8)
                    + Byte(C7 shr 8) + Byte(C8 shr 8) + Byte(C9 shr 8)) div 9;
                  B := (Byte(C1) + Byte(C2) + Byte(C3) + Byte(C4) +
                    Byte(C5) + Byte(C6) + Byte(C7) + Byte(C8) + Byte(C9)) div 9;

                  PByte(Pc)^ := PFSetARGB(PF, A, R, G, B);
                end;
              ifIndex8:
                begin
                  A := (Image.Palette[PByteArray(LineM1)[XM1]].A +
                    Image.Palette[PByteArray(LineM1)[X]].A +
                    Image.Palette[PByteArray(LineM1)[XP1]].A +
                    Image.Palette[PByteArray(Line)[XM1]].A +
                    Image.Palette[PByteArray(Line)[X]].A +
                    Image.Palette[PByteArray(Line)[XP1]].A +
                    Image.Palette[PByteArray(LineP1)[XM1]].A +
                    Image.Palette[PByteArray(LineP1)[X]].A +
                    Image.Palette[PByteArray(LineP1)[XP1]].A) div 9;
                  R := (Image.Palette[PByteArray(LineM1)[XM1]].R +
                    Image.Palette[PByteArray(LineM1)[X]].R +
                    Image.Palette[PByteArray(LineM1)[XP1]].R +
                    Image.Palette[PByteArray(Line)[XM1]].R +
                    Image.Palette[PByteArray(Line)[X]].R +
                    Image.Palette[PByteArray(Line)[XP1]].R +
                    Image.Palette[PByteArray(LineP1)[XM1]].R +
                    Image.Palette[PByteArray(LineP1)[X]].R +
                    Image.Palette[PByteArray(LineP1)[XP1]].R) div 9;
                  G := (Image.Palette[PByteArray(LineM1)[XM1]].G +
                    Image.Palette[PByteArray(LineM1)[X]].G +
                    Image.Palette[PByteArray(LineM1)[XP1]].G +
                    Image.Palette[PByteArray(Line)[XM1]].G +
                    Image.Palette[PByteArray(Line)[X]].G +
                    Image.Palette[PByteArray(Line)[XP1]].G +
                    Image.Palette[PByteArray(LineP1)[XM1]].G +
                    Image.Palette[PByteArray(LineP1)[X]].G +
                    Image.Palette[PByteArray(LineP1)[XP1]].G) div 9;
                  B := (Image.Palette[PByteArray(LineM1)[XM1]].B +
                    Image.Palette[PByteArray(LineM1)[X]].B +
                    Image.Palette[PByteArray(LineM1)[XP1]].B +
                    Image.Palette[PByteArray(Line)[XM1]].B +
                    Image.Palette[PByteArray(Line)[X]].B +
                    Image.Palette[PByteArray(Line)[XP1]].B +
                    Image.Palette[PByteArray(LineP1)[XM1]].B +
                    Image.Palette[PByteArray(LineP1)[X]].B +
                    Image.Palette[PByteArray(LineP1)[XP1]].B) div 9;

                  PByte(Pc)^ := ClosestColorInPal(Image.Palette, Image.PalSize,
                    ARGB(A, R, G, B));
                end;
            end;
          2: case Image.ImageFormat of
              ifIntensity16:
                begin
                  PWord(Pc)^ := (PWordArray(LineM1)[XM1] +
                    PWordArray(LineM1)[X] + PWordArray(LineM1)[XP1] +
                    PWordArray(Line)[XM1] +
                    PWordArray(Line)[X] + PWordArray(Line)[XP1] +
                    PWordArray(LineP1)[XM1] +
                    PWordArray(LineP1)[X] + PWordArray(LineP1)[XP1]) div 9;
                end;
              Low(if16bitARGBPF)..High(if16bitARGBPF):
                begin
                  C1 := PFGetARGB(PF, PWordArray(LineM1)[XM1]);
                  C2 := PFGetARGB(PF, PWordArray(LineM1)[X]);
                  C3 := PFGetARGB(PF, PWordArray(LineM1)[XP1]);
                  C4 := PFGetARGB(PF, PWordArray(Line)[XM1]);
                  C5 := PFGetARGB(PF, PWordArray(Line)[X]);
                  C6 := PFGetARGB(PF, PWordArray(Line)[XP1]);
                  C7 := PFGetARGB(PF, PWordArray(LineP1)[XM1]);
                  C8 := PFGetARGB(PF, PWordArray(LineP1)[X]);
                  C9 := PFGetARGB(PF, PWordArray(LineP1)[XP1]);

                  A := (Byte(C1 shr 24) + Byte(C2 shr 24) + Byte(C3 shr 24) +
                    Byte(C4 shr 24) + Byte(C5 shr 24) + Byte(C6 shr 24)
                    + Byte(C7 shr 24) + Byte(C8 shr 24) + Byte(C9 shr 24)) div
                    9;
                  R := (Byte(C1 shr 16) + Byte(C2 shr 16) + Byte(C3 shr 16) +
                    Byte(C4 shr 16) + Byte(C5 shr 16) + Byte(C6 shr 16)
                    + Byte(C7 shr 16) + Byte(C8 shr 16) + Byte(C9 shr 16)) div
                    9;
                  G := (Byte(C1 shr 8) + Byte(C2 shr 8) + Byte(C3 shr 8) +
                    Byte(C4 shr 8) + Byte(C5 shr 8) + Byte(C6 shr 8)
                    + Byte(C7 shr 8) + Byte(C8 shr 8) + Byte(C9 shr 8)) div 9;
                  B := (Byte(C1) + Byte(C2) + Byte(C3) + Byte(C4) +
                    Byte(C5) + Byte(C6) + Byte(C7) + Byte(C8) + Byte(C9)) div 9;

                  PWord(Pc)^ := PFSetARGB(PF, A, R, G, B);
                end;
            end;
          3:
            begin
              PRGBRec(Pc).R := (PRGBRecArray(LineM1)[XM1].R +
                PRGBRecArray(LineM1)[X].R + PRGBRecArray(LineM1)[XP1].R +
                PRGBRecArray(Line)[XM1].R +
                PRGBRecArray(Line)[X].R + PRGBRecArray(Line)[XP1].R +
                PRGBRecArray(LineP1)[XM1].R +
                PRGBRecArray(LineP1)[X].R + PRGBRecArray(LineP1)[XP1].R) div 9;
              PRGBRec(Pc).G := (PRGBRecArray(LineM1)[XM1].G +
                PRGBRecArray(LineM1)[X].G + PRGBRecArray(LineM1)[XP1].G +
                PRGBRecArray(Line)[XM1].G +
                PRGBRecArray(Line)[X].G + PRGBRecArray(Line)[XP1].G +
                PRGBRecArray(LineP1)[XM1].G +
                PRGBRecArray(LineP1)[X].G + PRGBRecArray(LineP1)[XP1].G) div 9;
              PRGBRec(Pc).B := (PRGBRecArray(LineM1)[XM1].B +
                PRGBRecArray(LineM1)[X].B + PRGBRecArray(LineM1)[XP1].B +
                PRGBRecArray(Line)[XM1].B +
                PRGBRecArray(Line)[X].B + PRGBRecArray(Line)[XP1].B +
                PRGBRecArray(LineP1)[XM1].B +
                PRGBRecArray(LineP1)[X].B + PRGBRecArray(LineP1)[XP1].B) div 9;
            end;
          4: case Image.ImageFormat of
              Low(if32bitARGB)..High(if32bitARGB):
                begin
                  PARGBRec(Pc).A := (PARGBRecArray(LineM1)[XM1].A +
                    PARGBRecArray(LineM1)[X].A + PARGBRecArray(LineM1)[XP1].A +
                    PARGBRecArray(Line)[XM1].A +
                    PARGBRecArray(Line)[X].A + PARGBRecArray(Line)[XP1].A +
                    PARGBRecArray(LineP1)[XM1].A +
                    PARGBRecArray(LineP1)[X].A + PARGBRecArray(LineP1)[XP1].A)
                    div 9;
                  PARGBRec(Pc).R := (PARGBRecArray(LineM1)[XM1].R +
                    PARGBRecArray(LineM1)[X].R + PARGBRecArray(LineM1)[XP1].R +
                    PARGBRecArray(Line)[XM1].R +
                    PARGBRecArray(Line)[X].R + PARGBRecArray(Line)[XP1].R +
                    PARGBRecArray(LineP1)[XM1].R +
                    PARGBRecArray(LineP1)[X].R + PARGBRecArray(LineP1)[XP1].R)
                    div 9;
                  PARGBRec(Pc).G := (PARGBRecArray(LineM1)[XM1].G +
                    PARGBRecArray(LineM1)[X].G + PARGBRecArray(LineM1)[XP1].G +
                    PARGBRecArray(Line)[XM1].G +
                    PARGBRecArray(Line)[X].G + PARGBRecArray(Line)[XP1].G +
                    PARGBRecArray(LineP1)[XM1].G +
                    PARGBRecArray(LineP1)[X].G + PARGBRecArray(LineP1)[XP1].G)
                    div 9;
                  PARGBRec(Pc).B := (PARGBRecArray(LineM1)[XM1].B +
                    PARGBRecArray(LineM1)[X].B + PARGBRecArray(LineM1)[XP1].B +
                    PARGBRecArray(Line)[XM1].B +
                    PARGBRecArray(Line)[X].B + PARGBRecArray(Line)[XP1].B +
                    PARGBRecArray(LineP1)[XM1].B +
                    PARGBRecArray(LineP1)[X].B + PARGBRecArray(LineP1)[XP1].B)
                    div 9;
                end;
              ifIntensity32:
                begin
                  Pc^ := (Int64(PULongArray(LineM1)[XM1]) +
                    Int64(PULongArray(LineM1)[X]) +
                    Int64(PULongArray(LineM1)[XP1]) +
                    Int64(PULongArray(Line)[XM1]) +
                    Int64(PULongArray(Line)[X]) + Int64(PULongArray(Line)[XP1])
                    + Int64(PULongArray(LineP1)[XM1]) +
                    Int64(PULongArray(LineP1)[X]) +
                    Int64(PULongArray(LineP1)[XP1])) div 9;
                end;
            end;
        end;
        Pc := Pointer(Long(Pc) + BytesPP);
      end;
    end;

    FreeMem(Image.Levels[I].Bits);
    Image.Levels[I].Bits := Data;
  end;
end;

procedure AddLevel(var Image: TImageData);
begin
  Inc(Image.LevelCount);
  ReallocMem(Image.Levels, Image.LevelCount * SizeOf(TImageLevel));
  FillChar(Image.Levels[Image.LevelCount - 1], SizeOf(TImageLevel), 0);
end;

procedure RemoveLevel(var Image: TImageData; Level: Long);
var
  I: Long;
begin
  if Level >= Image.LevelCount then
    Exit;

  FreeMemNil(Image.Levels[Level].Bits);
  for I := Level to Image.LevelCount - 2 do
    Image.Levels[I] := Image.Levels[I + 1];

  Dec(Image.LevelCount);
  ReallocMem(Image.Levels, Image.LevelCount * SizeOf(TImageLevel));
end;

function GetScanline(const Image: TImageData; Y, Level: Long): Pointer;
begin
  Result := nil;
  if Level >= Image.LevelCount then
    Exit;
  Result := Pointer(Long(Image.Levels[Level].Bits) + Y *
    Image.Levels[Level].Width * (Image.BitCount shr 3));
end;


{*************************************************************************}
{                      Color/PixelFormat  Functions                       }
{*************************************************************************}


function ARGB(A, R, G, B: Byte): TARGB;
begin
  Result := (A shl 24) or (R shl 16) or (G shl 8) or B;
end;

function RGB(R, G, B: Byte): TRGB;
begin
  Result := (R shl 16) or (G shl 8) or B;
end;

function BGR(R, G, B: Byte): TBGR;
begin
  Result := (B shl 16) or (G shl 8) or R;
end;

function GetA(ARGB: TARGB): Byte;
begin
  Result := ARGB shr 24;
end;

function GetR(ARGB: TARGB): Byte;
begin
  Result := ARGB shr 16;
end;

function GetG(ARGB: TARGB): Byte;
begin
  Result := ARGB shr 8;
end;

function GetB(ARGB: TARGB): Byte;
begin
  Result := ARGB;
end;

function ARGBRec(ARGB: TARGB): TARGBRec;
begin
  Result.Color := ARGB;
end;

function ARGBRec(A, R, G, B: Byte): TARGBRec;
begin
  Result.A := A;
  Result.R := R;
  Result.G := G;
  Result.B := B;
end;

function RGBRec(ARGB: TARGB): TRGBRec;
begin
  Result.B := ARGB;
  Result.G := ARGB shr 8;
  Result.R := ARGB shr 16;
end;

function RGBRec(R, G, B: Byte): TRGBRec;
begin
  Result.R := R;
  Result.G := G;
  Result.B := B;
end;

function BGRRec(ARGB: TARGB): TBGRRec;
begin
  Result.B := ARGB;
  Result.G := ARGB shr 8;
  Result.R := ARGB shr 16;
end;

procedure GetARGB(ARGB: TARGB; var A, R, G, B: Byte);
begin
  A := ARGB shr 24;
  R := ARGB shr 16;
  G := ARGB shr 8;
  B := ARGB;
end;

function SetA(ARGB: TARGB; A: Byte): TARGB;
begin
  Result := (A shl 24) or (Byte(ARGB shr 16) shl 16) or (Byte(ARGB shr 8) shl 8)
    or Byte(ARGB);
end;

function SetR(ARGB: TARGB; R: Byte): TARGB;
begin
  Result := (Byte(ARGB shr 24) shl 24) or (R shl 16) or (Byte(ARGB shr 8) shl 8)
    or Byte(ARGB);
end;

function SetG(ARGB: TARGB; G: Byte): TARGB;
begin
  Result := (Byte(ARGB shr 24) shl 24) or (Byte(ARGB shr 16) shl 16) or (G shl 8)
    or Byte(ARGB);
end;

function SetB(ARGB: TARGB; B: Byte): TARGB;
begin
  Result := (Byte(ARGB shr 24) shl 24) or (Byte(ARGB shr 16) shl 16) or
    (Byte(ARGB shr 8) shl 8) or B;
end;

function ARGBToStr(ARGB: TARGB): string;
begin
  ARGBTostrEx(ARGB, afHex);
end;

function ARGBToStrEx(ARGB: TARGB; Format: TARGBStrFmt): string;
begin
  Result := '';
  case Format of
    afHex: Result := SysUtils.Format('$%.8x', [ARGB]);
    afInt: Result := IntToStr(ARGB);
    afBytes: Result := SysUtils.Format('%d %d %d %d', [GetA(ARGB), GetR(ARGB),
        GetG(ARGB), GetB(ARGB)]);
    afDescBytes: Result := SysUtils.Format('A:%d R:%d G:%d B:%d', [GetA(ARGB),
        GetR(ARGB), GetG(ARGB), GetB(ARGB)]);
    afFullDesc: Result :=
      SysUtils.Format('Alpha: %d  Red: %d  Green: %d  Blue: %d', [GetA(ARGB),
        GetR(ARGB), GetG(ARGB), GetB(ARGB)]);
  end;
end;

function PixelFormat(ABitCount, RBitCount, GBitCount, BBitCount: Long):
  TPixelFormat;
begin
  Result.ABitMask := ((1 shl ABitCount) - 1) shl (RBitCount + GBitCount +
    BBitCount);
  Result.RBitMask := ((1 shl RBitCount) - 1) shl (GBitCount + BBitCount);
  Result.GBitMask := ((1 shl GBitCount) - 1) shl (BBitCount);
  Result.BBitMask := (1 shl BBitCount) - 1;
  Result.ABitCount := ABitCount;
  Result.RBitCount := RBitCount;
  Result.GBitCount := GBitCount;
  Result.BBitCount := BBitCount;
  {DONE -o Galfar : Here is problemo! Shift can be negative sometimes!! Abs not work}
  Result.AShift := RBitCount + GBitCount + BBitCount;
  Result.RShift := GBitCount + BBitCount;
  Result.GShift := BBitCount;
  Result.BShift := 0;
end;

function PixelFormatMask(ABitMask, RBitMask, GBitMask, BBitMask: Long):
  TPixelFormat;

  function GetBitCount(B: Long): Long;
  var
    I: Long;
  begin
    I := 0;
    while (I < 31) and (((1 shl I) and B) = 0) do
      Inc(I);
    Result := 0;
    while ((1 shl I) and B) <> 0 do
    begin
      Inc(I);
      Inc(Result);
    end;
  end;

begin
  Result := PixelFormat(GetBitCount(ABitMask), GetBitCount(RBitMask),
    GetBitCount(GBitMask),
    GetBitCount(BBitMask));
end;

function PFSetARGB(const PF: TPixelFormat; A, R, G, B: Byte): TARGB;
begin
  with PF do
    Result :=
      ((A shl ABitCount shr 8 shl AShift) and ABitMask) or
      ((R shl RBitCount shr 8 shl RShift) and RBitMask) or
      ((G shl GBitCount shr 8 shl GShift) and GBitMask) or
      ((B shl BBitCount shr 8 shl BShift) and BBitMask);
end;

procedure PFGetARGB(const PF: TPixelFormat; Color: TARGB; var A, R, G, B: Byte);
begin
  with PF do
  begin
    {TODO -o Galfar : Example : we have 1 bit Alpha and this is
    encoded (assume 8 bit mode) as 128. When we decode here
    we get 128 but before encoding it would be 255.
    It can be repaired by adding e.g. "or (A shl (8 - BitCount-1))".
    This should produce 192. This is good for alpha but not for colors.
    What to so with this?}
    A := (Color and ABitMask) shr AShift;
    A := A shl (8 - ABitCount);
    R := (Color and RBitMask) shr RShift;
    R := R shl (8 - RBitCount);
    G := (Color and GBitMask) shr GShift;
    G := G shl (8 - GBitCount);
    B := (Color and BBitMask) shl BShift;
    B := B shl (8 - BBitCount);
  end;
end;

function PFSetARGB(const PF: TPixelFormat; ARGB: TARGB): TARGB;
begin
  Result := PFSetARGB(PF, ARGB shr 24, ARGB shr 16, ARGB shr 8, ARGB);
end;

function PFGetARGB(const PF: TPixelFormat; Color: TARGB): TARGB;
var
  A, R, G, B: Byte;
begin
  PFGetARGB(PF, Color, A, R, G, B);
  Result := ARGB(A, R, G, B);
end;

function PFToStr(const PF: TPixelFormat): string;
begin
  Result := Format('AlphaBits: %d, RedBits: %d, GreenBits: %d, BlueBits: %d',
    [PF.ABitCount, PF.RBitCount, PF.GBitCount, PF.BBitCount]);
end;

function GreyScalePal(Entries: Long; Alpha: byte): PPalArray;
var
  Pal: PPalArray;
  I: Long;
begin
  GetMemFill(Pal, Entries * SizeOf(TARGBRec));
  for I := 0 to Entries - 1 do
    with Pal[I] do
    begin
      A := Alpha;
      R := I;
      G := I;
      B := I;
    end;
  Result := Pal;
end;

function HalftonePal(Entries, R, G, B: Long; Alpha: Byte): PPalArray;
var
  Pal: PPalArray;
  I: Long;
begin
  GetMemFill(Pal, Entries * SizeOf(TARGBRec));
  for I := 0 to Entries - 1 do
  begin
    Pal[I].A := Alpha;
    Pal[I].R := ((I shr (G + B - 1)) and (1 shl R - 1)) * 255 div (1 shl R - 1);
    Pal[I].G := ((I shr (B - 1)) and (1 shl G - 1)) * 255 div (1 shl G - 1);
    Pal[I].B := ((I shr 0) and (1 shl B - 1)) * 255 div (1 shl B - 1);
  end;
  Result := Pal;
end;

function RGBPalToARGBPal(const PalRGB: PPalRGBArray; Entries: Long; Alpha:
  byte): PPalArray;
var
  I: Long;
begin
  GetMem(Result, Entries * SizeOf(TARGBRec));
  for I := 0 to Entries - 1 do
    with Result[I] do
    begin
      A := Alpha;
      R := PalRGB[I].R;
      G := PalRGB[I].G;
      B := PalRGB[I].B;
    end;
end;


function ARGBPalToRGBPal(const PalARGB: PPalArray; Entries: Long): PPalRGBArray;
var
  I: Long;
begin
  GetMem(Result, Entries * SizeOf(TRGBRec));
  for I := 0 to Entries - 1 do
    with Result[I] do
    begin
      R := PalARGB[I].R;
      G := PalARGB[I].G;
      B := PalARGB[I].B;
    end;
end;

function RGBPalToARGBPal(const PalRGB: TPalRGB; Alpha: byte): TPalette;
var
  I: Long;
begin
  for I := 0 to 255 do
    with Result[I] do
    begin
      A := Alpha;
      R := PalRGB[I].R;
      G := PalRGB[I].G;
      B := PalRGB[I].B;
    end;
end;


function ARGBPalToRGBPal(const PalARGB: TPalette): TPalRGB;
var
  I: Long;
begin
  for I := 0 to 255 do
    with Result[I] do
    begin
      R := PalARGB[I].R;
      G := PalARGB[I].G;
      B := PalARGB[I].B;
    end;
end;


procedure PalSwapChannels(var PalARGB: PPalArray; Entries: Long);
var
  I: Long;
  T: Byte;
begin
  for I := 0 to Entries - 1 do
  begin
    T := PalARGB[I].R;
    PalARGB[I].R := PalARGB[I].B;
    PalARGB[I].B := T;
  end;
end;

function ClosestColorInPal(Pal: PPalArray; Entries: Long; Col: TARGB): Long;
var
  N: Long;
  PC: PARGBRec;
  C: TARGBRec;
  I, X, D: Long;
begin
  Result := 0;
  if Pal = nil then
    Exit;
  C := ARGBRec(Col);
  X := 765;
  N := 0;
  PC := Pointer(Pal);
  for I := 0 to Entries - 1 do
  begin
    if PC.A > C.A then
      D := PC.A - C.A
    else
      D := C.A - PC.A;
    if PC.R > C.R then
      Inc(D, PC.R - C.R)
    else
      Inc(D, C.R - PC.R);
    if PC.G > C.G then
      Inc(D, PC.G - C.G)
    else
      Inc(D, C.G - PC.G);
    if PC.B > C.B then
      Inc(D, PC.B - C.B)
    else
      Inc(D, C.B - PC.B);
    if D < X then
    begin
      X := D;
      N := I;
    end;
    Inc(PC);
  end;
  Result := N;
end;

function ImageFormatToPF(Format: TImageFormat): TPixelFormat;
begin
  case Format of
   // indexed modes
    ifIndex8..TImageFormat(31): Result := PixelFormat(8, 8, 8, 8);
   // intensity modes
    ifIntensity8: Result := PixelFormat(8, 0, 0, 0);
    // what here? 16 bpp for channel can not be used
    {TODO -o Galfar : Rewrite after pixel format renew - channels bigger than
     8 bit}
    ifIntensity16: Result := PixelFormat(4, 4, 4, 4);
    ifIntensity32: Result := PixelFormat(8, 8, 8, 8);
   // rgba modes
    ifR8G8B8, ifX8R8G8B8: Result := PixelFormat(0, 8, 8, 8);
    ifA8R8G8B8: Result := PixelFormat(8, 8, 8, 8);
    ifX1R5G5B5: Result := PixelFormat(0, 5, 5, 5);
    ifR5G6B5: Result := PixelFormat(0, 5, 6, 5);
    ifA1R5G5B5: Result := PixelFormat(1, 5, 5, 5);
    ifA4R4G4B4: Result := PixelFormat(4, 4, 4, 4);
    ifR3G3B2: Result := PixelFormat(0, 3, 3, 2);
    ifA1R2G3B2: Result := PixelFormat(1, 2, 3, 2);
  end;
end;

function ImageFormatBitCount(Format: TImageFormat): Long;
begin
  Result := 0;
  case Format of
   // indexed modes
    ifIndex8: Result := 8;
   // intensity modes
    ifIntensity8: Result := 8;
    ifIntensity16: Result := 16;
    ifIntensity32: Result := 32;
   // rgba modes
    ifR8G8B8: Result := 24;
    ifX8R8G8B8,
      ifA8R8G8B8: Result := 32;
    ifX1R5G5B5,
      ifR5G6B5,
      ifA1R5G5B5,
      ifA4R4G4B4: Result := 16;
    ifR3G3B2,
      ifA1R2G3B2: Result := 8;
  end;

end;

function ImageFormatToStr(Format: TImageFormat): string;
begin
  case Format of
   // indexed modes
    ifIndex8: Result := 'ifIndex8';
   // intensity modes
    ifIntensity8: Result := 'ifIntensity8';
    ifIntensity16: Result := 'ifIntensity16';
    ifIntensity32: Result := 'ifIntensity32';
   // rgba modes
    ifR8G8B8: Result := 'ifR8G8B8';
    ifX8R8G8B8: Result := 'ifX8R8G8B8';
    ifA8R8G8B8: Result := 'ifA8R8G8B8';
    ifX1R5G5B5: Result := 'ifX1R5G5B5';
    ifR5G6B5: Result := 'ifR5G6B5';
    ifA1R5G5B5: Result := 'ifA1R5G5B5';
    ifA4R4G4B4: Result := 'ifA4R4G4B4';
    ifR3G3B2: Result := 'ifR3G3B2';
    ifA1R2G3B2: Result := 'ifA1R2G3B2';
  end;
end;



end.

